r"""
One of the more popular integration schemes when dealing with orthogonal
polynomials are known as Gaussian quadrature. These are specially tailored
integration schemes each for different weighting schemes. Traditionally the
weights are given a form that does not adhere to the probability density
function rule of being normalized to 1, however the different is only scaling.

For example, consider the Gauss-Legendre which is optimized to perform the
integration:

.. math::
    \int_{-1}^1 g(x) dx \approx \sum_i W_i g(X_i)

The corresponding probability distribution that matches this contant weight
function on the :math:`(-1, 1)` interval, is ``chaospy.Uniform(-1, 1)``.
However, this distribution has a density of 0.5, instead of 1 in our example
here.

.. math::
    \int_{-1}^1 0.5 g(x) dx \approx \sum_i W_i g(X_i)

So to use ``chaospy`` to create a true Gaussian quadrature rule, one often has
to multiply the weights :math:`W_i` with some adjustment scalar. For example::

    >>> distribution = chaospy.Uniform(-1, 1)
    >>> N = 3
    >>> adjust_scalar = 2
    >>> X, W = chaospy.generate_quadrature(N, distribution, rule="gaussian")
    >>> W *= adjust_scalar
    >>> print(X)
    [[-0.86113631 -0.33998104  0.33998104  0.86113631]]
    >>> print(W)
    [0.34785485 0.65214515 0.65214515 0.34785485]

Here ``rule="gaussian"`` is the flag that indicate that Gaussian quadrature
should be used.

The various constants and distributions to achieve the various quadrature rules
are as follows.

==================== ======================= ========================= ===================
Scheme               Weight function         Distribution              Adjustment
==================== ======================= ========================= ===================
Hermite              :math:`e^{-x^2}`        ``Normal(0, 2**-0.5)``    :math:`\sqrt{\pi}`
Legendre             :math:`1`               ``Uniform(-1, 1)``        :math:`2`
Jakobi               :math:`(1-x)^a(1+x)^b`  ``Beta(a+1, b+1, -1, 1)`` :math:`2^{a+b}`
1. order Chebyshev   :math:`1/\sqrt{1-x^2}`  ``Beta(0.5, 0.5, -1, 1)`` :math:`1/2`
2. order Chebyshev   :math:`\sqrt{1-x^2}`    ``Beta(1.5, 1.5, -1, 1)`` :math:`2`
Laguerre             :math:`e^{-x}`          ``Exponential()``         :math:`1`
Generalized Laguerre :math:`x^a e^{-x}`      ``Gamma(a+1)``            :math:`\Gamma(a+1)`
Gegenbaur            :math:`(1-x^2)^{a-0.5}` ``Beta(a+.5,a+.5,-1,1)``  :math:`2^{2a-1}`
==================== ======================= ========================= ===================

However, the list is not limited to these cases. Any and all valid weight
function are supported this way. However, not all weight functions does not
work very well. E.g. using the log-normal probability density function as
a weight function is known to scale badly. Which one works or not, depends on
context, so any non-standard use has to be done with some care.
"""
from .frontend import (
    construct_recurrence_coefficients, RECURRENCE_ALGORITHMS)
from .jacobi import coefficients_to_quadrature

from .chebyshev import modified_chebyshev
from .lanczos import lanczos
from .stieltjes import discretized_stieltjes, analytical_stieljes



"""
The script is adapted from the routine RKPW in W.B. Gragg and W.J. Harrod,
"The numerically stable reconstruction of Jacobi matrices from spectral data",
Numer. Math. 44 (1984), 317-335.

Example usage
-------------

Basic usage::

    >>> dist = chaospy.Beta(3, 6)
    >>> abscissas, weights = chaospy.generate_quadrature(
    ...     10, dist, rule="clenshaw_curtis")
    >>> alpha, beta = lanczos(3, abscissas, weights)
    >>> print(alpha)
    [0.33333333 0.39393939 0.42680399 0.44727847]
    >>> print(beta)
    [1.         0.02222222 0.03471074 0.04207915]
"""
import numpy

import chaospy.poly.base
import chaospy.poly.collection


def lanczos(order, abscissas, weights):

    abscissas = abscissas.astype(float)
    weights = weights.astype(float)
    assert len(weights.shape) == 1
    assert len(abscissas.shape) == 2
    assert abscissas.shape[-1] == len(weights)

    assert len(weights) > order
    alpha = abscissas[:, :order+1].copy()
    beta = numpy.zeros((len(abscissas), order+1))
    beta[:, 0] = weights[0]

    for idx in range(1, len(weights)):

        gamma = 1
        increment_new = 0

        for idy in range(min(idx, order+1)):

            increment = increment_new
            beta_new = gamma*(beta[:, idy]+weights[idx])
            sigma = 1-gamma
            gamma = numpy.where(
                beta[:, idy] <= -weights[idx], 1,
                beta[:, idy]/(beta[:, idy]+weights[idx]))
            increment_new = (
                (1-gamma)*(alpha[:, idy]-abscissas[:, idx])-gamma*increment)
            weights[idx] = numpy.where(
                gamma >= 1, sigma*beta[:, idy], increment_new**2/(1-gamma))

            alpha[:, idy] -= increment_new-increment
            beta[:, idy] = beta_new

    return numpy.asfarray([alpha.flatten(), beta.flatten()])



"""Collection of mathematical operators."""
from .addition import Add, add
from .multiply import Mul, mul
from .negative import Neg, neg
from .power import Pow

from .joint import J
from .iid import Iid

from .trunkation import Trunc, Trunk, trunk

from .logarithm import Log
from .logarithm10 import Log10
from .logarithmn import Logn

from .sin import Sin
from .arcsin import Arcsin
from .cos import Cos
from .arccos import Arccos
from .tan import Tan
from .arctan import Arctan
from .sinh import Sinh
from .arcsinh import Arcsinh
from .cosh import Cosh
from .arccosh import Arccosh
from .tanh import Tanh
from .arctanh import Arctanh



"""
Evaluate forward Rosenblatt transformation.

In the case of univariate distributions, this is equivalent to the cumulative
distribution function (CDF).

Example usage
-------------

Define a simple distribution and data::

    >>> class Exponential(chaospy.Dist):
    ...     def _cdf(self, x_data, alpha): return 1-numpy.e**(-alpha*x_data)
    >>> dist = Exponential(alpha=2)
    >>> x_data = numpy.array([[0.1, 0.2, 0.3]])

Normal usage::

    >>> print(numpy.around(evaluate_forward(dist, x_data), 4))
    [[0.1813 0.3297 0.4512]]

Use non-default parameters::

    >>> print(numpy.around(evaluate_forward(
    ...     dist, x_data, parameters={"alpha": 1}), 4))
    [[0.0952 0.1813 0.2592]]
"""
import numpy
from .parameters import load_parameters


def evaluate_forward(
        distribution,
        x_data,
        parameters=None,
        cache=None,
):
    """
    Evaluate forward Rosenblatt transformation.

    Args:
        distribution (Dist):
            Distribution to evaluate.
        x_data (numpy.ndarray):
            Locations for where evaluate forward transformation at.
        parameters (:py:data:typing.Any):
            Collection of parameters to override the default ones in the
            distribution.
        cache (:py:data:typing.Any):
            A collection of previous calculations in case the same distribution
            turns up on more than one occasion.

    Returns:
        The cumulative distribution values of ``distribution`` at location
        ``x_data`` using parameters ``parameters``.
    """
    assert len(x_data) == len(distribution), (
        "distribution %s is not of length %d" % (distribution, len(x_data)))
    assert hasattr(distribution, "_cdf"), (
        "distribution require the `_cdf` method to function.")

    cache = cache if cache is not None else {}

    parameters = load_parameters(
        distribution, "_cdf", parameters=parameters, cache=cache)

    # Store cache.
    cache[distribution] = x_data

    # Evaluate forward function.
    out = numpy.zeros(x_data.shape)
    out[:] = distribution._cdf(x_data, **parameters)

    return out



"""Clayton copula."""
import numpy
from scipy import special

from .baseclass import Archimedean, Copula
from ..baseclass import Dist

class clayton(Archimedean):
    """Clayton copula."""

    def __init__(self, length, theta=1., eps=1e-6):
        self.length = length
        Dist.__init__(self, th=float(theta), eps=eps)

    def __len__(self):
        return self.length

    def gen(self, x, th):
        return (x**-th-1)/th

    def igen(self, x, th):
        return (1.+th*x)**(-1./th)


class Clayton(Copula):
    """
    Clayton Copula.

    Args:
        dist (Dist) : The Distribution to wrap
        theta (float) : Copula parameter

    Returns:
        (Dist) : The resulting copula distribution.

    Examples:
        >>> distribution = chaospy.Clayton(
        ...     chaospy.Iid(chaospy.Uniform(), 2), theta=2)
        >>> print(distribution)
        Clayton(Iid(Uniform(lower=0, upper=1), 2), theta=2)
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(distribution.inv(mesh), 4))
        [[[0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]]
        <BLANKLINE>
         [[0.1987 0.3758 0.5197]
          [0.3101 0.5464 0.6994]
          [0.4777 0.7361 0.8525]]]
        >>> print(numpy.around(distribution.fwd(distribution.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(distribution.pdf(distribution.inv(mesh)), 4))
        [[2.3697 1.4016 1.1925]
         [1.9803 1.4482 1.5538]
         [1.0651 1.1642 1.6861]]
        >>> print(numpy.around(distribution.sample(4), 4))
        [[0.6536 0.115  0.9503 0.4822]
         [0.9043 0.0852 0.3288 0.4633]]
        >>> print(numpy.around(distribution.mom((1, 2)), 4))
        0.2196
    """

    def __init__(self, dist, theta=2., eps=1e-6):
        """
        Args:
            dist (Dist) : The Distribution to wrap
            theta (float) : Copula parameter
        """
        self._repr = {"theta": theta}
        trans = clayton(len(dist), theta=theta, eps=eps)
        return Copula.__init__(self, dist=dist, trans=trans)



"""Truncated exponential distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class truncexpon(Dist):
    """Truncated exponential distribution."""

    def __init__(self, b):
        Dist.__init__(self, b=b)

    def _pdf(self, x, b):
        return numpy.exp(-x)/(1-numpy.exp(-b))

    def _cdf(self, x, b):
        return (1.0-numpy.exp(-x))/(1-numpy.exp(-b))

    def _ppf(self, q, b):
        return -numpy.log(1-q+q*numpy.exp(-b))

    def _bnd(self, x, b):
        return 0.0, b


class TruncExponential(Add):
    """
    Truncated exponential distribution.

    Args:
        upper (float, Dist) : Location of upper threshold
        scale (float, Dist) : Scaling parameter in the exponential distribution
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.TruncExponential(2, 4)
        >>> print(distribution)
        TruncExponential(scale=4, shift=0, upper=2)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.     0.4142 0.8763 1.3988 2.    ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.6354 0.5729 0.5104 0.4479 0.3854]
        >>> print(numpy.around(distribution.sample(4), 4))
        [1.1891 0.1852 1.873  0.8415]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.917
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[1.0163 1.0024 1.0008]
         [0.3292 0.2671 0.2572]]
    """

    def __init__(self, upper=1, scale=1, shift=0):
        self._repr = {"upper": upper, "scale": scale, "shift": shift}
        Add.__init__(
            self, left=truncexpon((upper-shift)*1./scale)*scale, right=shift)


Truncexpon = deprecation_warning(TruncExponential, "Truncexpon")



"""Binomial probability distribution."""
from functools import wraps
import numpy
from scipy import special

from ..baseclass import Dist


class Binomial(Dist):
    """
    Binomial probability distribution.

    Point density:
        comb(N, x) p^x (1-p)^{N-x}      x in {0, 1, ..., N}

    Examples:
        >>> distribution = chaospy.Binomial(5, 0.5)
        >>> print(distribution)
        Binomial(prob=0.5, size=5)
        >>> q = numpy.linspace(0, 1, 8)
        >>> print(numpy.around(distribution.inv(q), 4))
        [0 1 2 2 3 3 4 5]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.0312 0.1875 0.5    0.5    0.8125 0.8125 0.9688 1.    ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.0312 0.1562 0.3125 0.3125 0.3125 0.3125 0.1562 0.0312]
        >>> print(numpy.around(distribution.sample(10), 4))
        [3 1 4 2 4 2 1 2 2 4]
        >>> print(numpy.around(distribution.mom([1, 2, 3]), 4))
        [ 2.5  7.5 25. ]
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[2.5  2.5  2.5 ]
         [1.25 2.   2.25]]
    """
    interpret_as_integer = True

    def __init__(self, size, prob):
        Dist.__init__(self, size=size, prob=prob)

    def _cdf(self, x_data, size, prob):
        return special.bdtr(numpy.floor(x_data), numpy.floor(size), prob)

    def _ppf(self, q_data, size, prob):
        return numpy.ceil(special.bdtrik(q_data, numpy.floor(size), prob))

    def _pdf(self, x_data, size, prob):
        return special.comb(size, x_data)*prob**x_data*(1-prob)**(size-x_data)

    def _bnd(self, x_data, size, prob):
        return 0, numpy.floor(size)+1

    def _mom(self, k_data, size, prob):
        x_data = numpy.arange(int(size)+1, dtype=int)
        return numpy.sum(x_data**k_data*self._pdf(
            x_data, size=numpy.floor(size), prob=prob))

    def _ttr(self, k_data, size, prob):
        """Krawtchouk rule"""
        from chaospy.quadrature import discretized_stieltjes
        abscissas = numpy.arange(0, numpy.floor(size)+1)
        weights = self._pdf(abscissas, size, prob)
        (alpha, beta), _, _ = discretized_stieltjes(k_data, [abscissas], weights)
        return alpha[0, -1], beta[0, -1]



"""Generalized gamma distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class generalized_gamma(Dist):
    """Generalized gamma distribution."""

    def __init__(self, a, c):
        Dist.__init__(self, a=a, c=c)

    def _pdf(self, x, a, c):
        return abs(c)* numpy.exp((c*a-1)*numpy.log(x)-x**c- special.gammaln(a))

    def _cdf(self, x, a, c):
        val = special.gammainc(a, x**c)
        cond = c + 0*val
        return numpy.where(cond > 0, val, 1-val)

    def _ppf(self, q, a, c):
        val1 = special.gammaincinv(a, q)
        val2 = special.gammaincinv(a, 1.0-q)
        ic = 1.0/c
        cond = c+0*val1
        return numpy.where(cond > 0, val1**ic, val2**ic)

    def _mom(self, k, a, c):
        return special.gamma((c+k)*1./a)/special.gamma(c*1./a)

    def _bnd(self, x, a, c):
        return 0., self._ppf(1-1e-10, a, c)


class GeneralizedGamma(Add):
    """
    Generalized gamma distribution

    Args:
        shape1 (float, Dist) : Shape parameter 1
        shape2 (float, Dist) : Shape parameter 2
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.GeneralizedGamma(3, 2, 2, 2)
        >>> print(distribution)
        GeneralizedGamma(scale=2, shape1=3, shape2=2, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [4.4779 5.0233 5.5244 6.1372]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.3145 0.4016 0.3807 0.2624]
        >>> print(numpy.around(distribution.sample(4), 4))
        [5.6691 4.1674 7.0214 5.2264]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.477
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[5.6341 5.9361 6.2271]
         [0.9553 1.8381 2.6689]]
    """

    def __init__(self, shape1, shape2, scale, shift):
        self._repr = {
            "shape1": shape1, "shape2": shape2, "scale": scale, "shift": shift}
        Add.__init__(
            self, left=generalized_gamma(shape1, shape2)*scale, right=shift)


Gengamma = deprecation_warning(GeneralizedGamma, "Gengamma")



"""Exponential power distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class exponential_power(Dist):
    """Exponential power distribution."""

    def __init__(self, b=1):
        Dist.__init__(self, b=b)

    def _pdf(self, x, b):
        xbm1 = x**(b-1.0)
        xb = xbm1 * x
        return numpy.exp(1)*b*xbm1 * numpy.exp(xb - numpy.exp(xb))

    def _cdf(self, x, b):
        xb = x**b
        return -numpy.expm1(-numpy.expm1(xb))

    def _ppf(self, q, b):
        return pow(numpy.log1p(-numpy.log1p(-q)), 1.0/b)

    def _bnd(self, x, b):
        return 0, self._ppf(1-1e-10, b)


class ExponentialPower(Add):
    """
    Exponential power distribution.

    Also known as Generalized error distribution and Generalized normal
    distribution version 1.

    Args:
        shape (float, Dist): Shape parameter
        scale (float, Dist): Scaling parameter
        shift (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.ExponentialPower(2, 2, 1)
        >>> print(distribution)
        ExponentialPower(scale=2, shape=2, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [1.8976 2.2848 2.6129 2.9587]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.4392 0.5823 0.6182 0.5111]
        >>> print(numpy.around(distribution.sample(4), 4))
        [2.7003 1.679  3.3551 2.4223]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.4314
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[2.3851 2.4491 2.5043]
         [0.3363 0.4539 0.5185]]
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=exponential_power(shape)*scale, right=shift)


Exponpow = deprecation_warning(ExponentialPower, "Exponpow")



"""Tangent."""
import numpy

from ..baseclass import Dist
from .. import evaluation, approximation


class Tan(Dist):
    """
    Tangent.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Tan(chaospy.Uniform(0, 1))
        >>> print(distribution)
        Tan(Uniform(lower=0, upper=1))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.2027 0.4228 0.6841 1.0296]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.9605 0.8484 0.6812 0.4854]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.7659 0.1155 1.3992 0.5234]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.6156
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[0.6156 0.8079 0.7829]
         [1.     0.1784 0.1654]]
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        """Probability density function."""
        return evaluation.evaluate_density(
            dist, numpy.arctan(x), cache=cache)/(1+x*x)

    def _cdf(self, x, dist, cache):
        return evaluation.evaluate_forward(dist, numpy.arctan(x))

    def _ppf(self, q, dist, cache):
        return numpy.tan(evaluation.evaluate_inverse(
            dist, q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.tan(evaluation.evaluate_bound(
            dist, numpy.arctan(x), cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"




"""Construct recurrence coefficients."""
import numpy

from .chebyshev import modified_chebyshev
from .jacobi import coefficients_to_quadrature
from .lanczos import lanczos
from .stieltjes import discretized_stieltjes

RECURRENCE_ALGORITHMS = ("analytical", "chebyshev", "lanczos", "stieltjes")


def construct_recurrence_coefficients(
        order,
        dist,
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    """
    Frontend wrapper for constructing *three terms recurrence* coefficients.

    The algorithm for constructing recurrence coefficients can be specified
    using the ``recurrence_algorithm`` flag. It accepts the strings:

    ``analytical``
        Some distributions have a built-in method for generating three terms
        recurrence coefficients. If that is not the case, raise an appropriate
        error. The most stable method, if available.
    ``chebyshev``
        Use modified Chebyshev algorithm to convert raw statistical moments to
        the recurrence coefficients. A good algorithm for when raw statistical
        moments are known analytically.
    ``lanczos``
        Use a known relationship between the Jakobi matrix and a matrix
        consisting of abscissas and weights from an alternative integration
        scheme to estimate the recurrence coefficients. Stabilized using ideas
        by Rutishauser. An alternative method to ``stieltjes``.
    ``stieltjes``
        Use the discretized Stieltjes algorithm for iterative estimate each
        recurrence coefficient using numerical integration. Typically the
        method known to have the highest convergence rate when no analytical
        information is available.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Dist):
            The distribution which density will be used as weight function.
            Assumed to one-dimensional.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        accuracy (int):
            In the case ``rule`` is used, defines the quadrature order of the
            scheme used. In practice, must be at least as large as ``order``.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.

    Returns:
        (typing.List[numpy.ndarray]):
            List of recurrence coefficients with shape ``(2, order+1)``. The
            alpha and beta coefficients can found in ``out[0]`` and ``out[1]``
            respectively.

    Examples:
        >>> distribution = chaospy.Normal(0, 1)
        >>> coefficients = chaospy.construct_recurrence_coefficients(
        ...     4, distribution, recurrence_algorithm="stieltjes")
        >>> print(numpy.around(coefficients, 4))
        [[[-0.  0. -0.  0. -0.]
          [ 1.  1.  2.  3.  4.]]]
        >>> distribution = chaospy.J(chaospy.Exponential(), chaospy.Uniform())
        >>> coefficients = chaospy.construct_recurrence_coefficients(
        ...     [2, 4], distribution, recurrence_algorithm="chebyshev")
        >>> print(numpy.around(coefficients[0], 4))
        [[1. 3. 5.]
         [1. 1. 4.]]
        >>> print(numpy.around(coefficients[1], 4))
        [[0.5    0.5    0.5    0.5    0.5   ]
         [1.     0.0833 0.0667 0.0643 0.0635]]
    """
    if not recurrence_algorithm:
        try:
            return construct_recurrence_coefficients(
                order, dist, rule, accuracy, recurrence_algorithm="analytical")
        except NotImplementedError:
            return construct_recurrence_coefficients(
                order, dist, rule, accuracy, recurrence_algorithm="stieltjes")

    if len(dist) > 1:
        orders = (order*numpy.ones(len(dist), dtype=int)).tolist()
        return [construct_recurrence_coefficients(
            int(order_), dist_, rule, accuracy, recurrence_algorithm)[0]
                for dist_, order_ in zip(dist, orders)]

    assert recurrence_algorithm in RECURRENCE_ALGORITHMS, (
        "recurrence algorithm '%s' not recognized" % recurrence_algorithm)
    assert not rule.startswith("gauss"), (
        "recursive Gaussian quadrature construct")

    if recurrence_algorithm == "analytical":
        coeffs = dist.ttr(numpy.arange(order+1, dtype=int))

    elif recurrence_algorithm == "chebyshev":
        moments = dist.mom(numpy.arange(2*(order+1), dtype=int))
        coeffs = modified_chebyshev(moments)

    elif recurrence_algorithm == "lanczos":
        from ..frontend import generate_quadrature
        abscissas, weights = generate_quadrature(accuracy, dist, rule=rule)
        coeffs = lanczos(order, abscissas, weights)

    elif recurrence_algorithm == "stieltjes":
        from ..frontend import generate_quadrature
        abscissas, weights = generate_quadrature(accuracy, dist, rule=rule)
        coeffs, _, _ = discretized_stieltjes(order, abscissas, weights)

    return [coeffs.reshape(2, int(order)+1)]



"""
Lagrange polynomials are not a method for creating orthogonal polynomials.
Instead it is an interpolation method for creating an polynomial expansion that
has the property that each polynomial interpolates exactly one point in space
with the value 1 and has the value 0 for all other interpolation values.
For more details, see this `article on Lagrange polynomials`_.

.. _article on Lagrange polynomials: https://en.wikipedia.org/wiki/Lagrange_polynomial
"""
import numpy
import chaospy

def lagrange_polynomial(abscissas, sort="G"):
    """
    Create Lagrange polynomials.

    Args:
        abscissas (numpy.ndarray):
            Sample points where the Lagrange polynomials shall be defined.

    Example:
        >>> print(chaospy.around(chaospy.lagrange_polynomial([-10, 10]), 4))
        [-0.05q0+0.5, 0.05q0+0.5]
        >>> print(chaospy.around(chaospy.lagrange_polynomial([-1, 0, 1]), 4))
        [0.5q0^2-0.5q0, -q0^2+1.0, 0.5q0^2+0.5q0]
        >>> poly = chaospy.lagrange_polynomial([[1, 0, 1], [0, 1, 2]])
        >>> print(chaospy.around(poly, 4))
        [0.5q0-0.5q1+0.5, -q0+1.0, 0.5q0+0.5q1-0.5]
        >>> print(numpy.around(poly([1, 0, 1], [0, 1, 2]), 4))
        [[1. 0. 0.]
         [0. 1. 0.]
         [0. 0. 1.]]
        >>> nodes = numpy.array([[ 0.17,  0.15,  0.17,  0.19],
        ...                      [14.94, 16.69, 16.69, 16.69]])
        >>> poly = chaospy.lagrange_polynomial(nodes)  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        LinAlgError: Lagrange abscissas resulted in invertible matrix
    """
    abscissas = numpy.asfarray(abscissas)
    if len(abscissas.shape) == 1:
        abscissas = abscissas.reshape(1, abscissas.size)
    dim, size = abscissas.shape

    order = 1
    while chaospy.bertran.terms(order, dim) < size:
        order += 1

    indices = numpy.array(chaospy.bertran.bindex(0, order, dim, sort)[:size])
    idx, idy = numpy.mgrid[:size, :size]

    matrix = numpy.prod(abscissas.T[idx]**indices[idy], -1)
    det = numpy.linalg.det(matrix)
    if det == 0:
        raise numpy.linalg.LinAlgError(
            "Lagrange abscissas resulted in invertible matrix")

    vec = chaospy.poly.basis(0, order, dim, sort)[:size]

    coeffs = numpy.zeros((size, size))

    if size == 1:
        out = chaospy.poly.basis(0, 0, dim, sort)*abscissas.item()

    elif size == 2:
        coeffs = numpy.linalg.inv(matrix)
        out = chaospy.poly.sum(vec*(coeffs.T), 1)

    else:
        
        for i in range(size):
            if i%2 != 0:
                k = 1
            else:
                k=0
            for j in range(size):
                if k%2 == 0:
                    coeffs[i, j] += numpy.linalg.det(matrix[1:, 1:])
                else:
                    if size%2 == 0:
                        coeffs[i, j] += -numpy.linalg.det(matrix[1:, 1:])
                    else:
                        coeffs[i, j] += numpy.linalg.det(matrix[1:, 1:])
                matrix = numpy.roll(matrix, -1, axis=0) 
                k += 1
            matrix = numpy.roll(matrix, -1, axis=1) 
        coeffs /= det
        out = chaospy.poly.sum(vec*(coeffs.T), 1)

    return out



"""Fisk or Log-logistic distribution."""
from ..baseclass import Dist
from ..operators.addition import Add


class fisk(Dist):
    """Fisk or Log-logistic distribution."""

    def __init__(self, c=1.):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        output = c*x**(c-1.)
        output /= (1+x**c)**2
        return output

    def _cdf(self, x, c):
        return 1./(1+x**-c)

    def _ppf(self, q, c):
        return (q**(-1.0)-1)**(-1.0/c)

    def _bnd(self, x, c):
        return 0, self._ppf(1-1e-10, c)


class Fisk(Add):
    """
    Fisk or Log-logistic distribution.

    Args:
        shape (float, Dist): Shape parameter
        scale (float, Dist): Scaling parameter
        shift (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.Fisk(3, 2, 1)
        >>> print(distribution)
        Fisk(scale=2, shape=3, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.2599 2.7472 3.2894 4.1748]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.381  0.4121 0.3145 0.1512]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.4714 2.013  6.3474 2.9531]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.4184
    """
    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=fisk(shape)*scale, right=shift)



"""Alpha probability distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class alpha(Dist):
    """Standard Alpha distribution."""

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _cdf(self, x, a):
        return special.ndtr(a-1./x) / special.ndtr(a)

    def _ppf(self, q, a):
        return 1.0/(a-special.ndtri(q*special.ndtr(a)))

    def _pdf(self, x, a):
        return 1.0/(x**2)/special.ndtr(a)*numpy.e**(.5*(a-1.0/x)**2)/numpy.sqrt(2*numpy.pi)

    def _bnd(self, x, a):
        return 0, self._ppf(1-1e-10, a)


class Alpha(Add):
    """
    Alpha distribution.

    Args:
        shape (float, Dist): Shape parameter
        scale (float, Dist): Scale Parameter
        shift (float, Dist): Location of lower threshold

    Examples:
        >>> distribution = chaospy.Alpha(2, 0.5, 4)
        >>> print(distribution)
        Alpha(scale=0.5, shape=2, shift=4)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [4.1676 4.2039 4.2465 4.3104 4.4521]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [11.7723  5.4345  3.361   2.2848  1.4892]
        >>> print(numpy.around(distribution.sample(4), 4))
        [4.304  4.1556 4.9362 4.2413]
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=alpha(shape)*scale, right=shift)



"""
Constructing a multivariate probability distribution can be done using the
:func:`~chaospy.distributions.operators.joint.J` constructor. E.g.::

    >>> distribution = chaospy.J(
    ...     chaospy.Normal(0, 1), chaospy.Uniform(0, 1))

The created multivariate distribution behaves much like the univariate case::

    >>> mesh = numpy.meshgrid(
    ...     numpy.linspace(0.25, 0.75, 3),
    ...     numpy.linspace(0.25, 0.75, 3),
    ... )
    >>> print(numpy.around(distribution.cdf(mesh), 4))
    [[0.1497 0.1729 0.1933]
     [0.2994 0.3457 0.3867]
     [0.449  0.5186 0.58  ]]
    >>> print(numpy.around(distribution.pdf(mesh), 4))
    [[0.3867 0.3521 0.3011]
     [0.3867 0.3521 0.3011]
     [0.3867 0.3521 0.3011]]
    >>> print(numpy.around(distribution.sample(
    ...     size=6, rule="H", antithetic=True), 4))
    [[-1.1503  1.1503 -1.1503  1.1503  0.3186 -0.3186]
     [ 0.4444  0.4444  0.5556  0.5556  0.7778  0.7778]]
    >>> print(distribution.mom([[2, 4, 6], [1, 2, 3]]))
    [0.5  1.   3.75]
"""
import numpy

from ..baseclass import Dist, StochasticallyDependentError
from .. import evaluation


class J(Dist):
    """
    Joint random variable generator.

    Args:
        args: Dist
            Distribution to join together.
    """

    def __init__(self, *args):
        args = [dist for arg in args
                for dist in (arg if isinstance(arg, J) else [arg])]
        assert all(isinstance(dist, Dist) for dist in args)
        self.indices = [0] + numpy.cumsum([len(dist) for dist in args[:-1]]).tolist()
        self.inverse_map = {dist: idx for idx, dist in zip(self.indices, args)}
        prm = {"_%03d" % idx: dist for idx, dist in zip(self.indices, args)}
        Dist.__init__(self, **prm)

    def _cdf(self, xloc, cache, **kwargs):
        """
        Examples:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> print(dist.fwd([[-0.5, 0.5, 1.5], [-1, 0, 1]]))
            [[0.         0.5        1.        ]
             [0.15865525 0.5        0.84134475]]
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> print(dist.fwd([[-0.5, 0.5, 1.5], [0, 1, 2]]))
            [[0.  0.5 1. ]
             [0.5 0.5 0.5]]
        """
        uloc = numpy.zeros(xloc.shape)
        for dist in evaluation.sorted_dependencies(self, reverse=True):
            if dist not in self.inverse_map:
                continue
            idx = self.inverse_map[dist]
            uloc[idx:idx+len(dist)] = evaluation.evaluate_forward(
                dist, xloc[idx:idx+len(dist)], cache=cache)[0]
        assert uloc.shape == xloc.shape
        return uloc

    def _range(self, xloc, cache):
        """
        Special handle for finding bounds on constrained dists.

        Example:
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> print(dist.range())
            [[0. 0.]
             [1. 2.]]
        """
        uloc = numpy.zeros((2, len(self)))
        for dist in evaluation.sorted_dependencies(self, reverse=True):
            if dist not in self.inverse_map:
                continue
            idx = self.inverse_map[dist]
            (uloc[0, idx:idx+len(dist)],
             uloc[1, idx:idx+len(dist)]) = evaluation.evaluate_bound(
                 dist, xloc[idx:idx+len(dist)], cache=cache)
        return uloc

    def _bnd(self, xloc, cache, **kwargs):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> print(dist.range([[-0.5, 0.5, 1.5], [-1, 0, 1]]))
            [[[ 0.   0.   0. ]
              [-7.5 -7.5 -7.5]]
            <BLANKLINE>
             [[ 1.   1.   1. ]
              [ 7.5  7.5  7.5]]]
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> print(dist.range([[-0.5, 0.5, 1.5], [0, 1, 2]]))
            [[[0. 0. 0.]
              [0. 0. 0.]]
            <BLANKLINE>
             [[1. 1. 1.]
              [2. 2. 2.]]]
        """
        uloc = numpy.zeros((2,)+xloc.shape)
        for dist in evaluation.sorted_dependencies(self, reverse=True):
            if dist not in self.inverse_map:
                continue
            idx = self.inverse_map[dist]
            uloc[:, idx:idx+len(dist)] = evaluation.evaluate_bound(
                dist, xloc[idx:idx+len(dist)], cache=cache)
        return uloc

    def _pdf(self, xloc, cache, **kwargs):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> print(numpy.around(dist.pdf([[-0.5, 0.5, 1.5], [-1, 0, 1]]), 4))
            [0.     0.3989 0.    ]
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> print(dist.pdf([[-0.5, 0.5, 1.5], [0, 1, 2]]))
            [0. 1. 0.]
        """
        floc = numpy.zeros(xloc.shape)
        for dist in evaluation.sorted_dependencies(self, reverse=True):
            if dist not in self.inverse_map:
                continue
            idx = self.inverse_map[dist]
            floc[idx:idx+len(dist)] = evaluation.evaluate_density(
                dist, xloc[idx:idx+len(dist)], cache=cache)[0]
        return floc

    def _ppf(self, qloc, cache, **kwargs):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> print(numpy.around(dist.inv([[0.1, 0.2, 0.3], [0.3, 0.3, 0.4]]), 4))
            [[ 0.1     0.2     0.3   ]
             [-0.5244 -0.5244 -0.2533]]
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> print(numpy.around(dist.inv([[0.1, 0.2, 0.3], [0.3, 0.3, 0.4]]), 4))
            [[0.1 0.2 0.3]
             [0.4 0.5 0.7]]
        """
        xloc = numpy.zeros(qloc.shape)
        for dist in evaluation.sorted_dependencies(self, reverse=True):
            if dist not in self.inverse_map:
                continue
            idx = self.inverse_map[dist]
            xloc[idx:idx+len(dist)] = evaluation.evaluate_inverse(
                dist, qloc[idx:idx+len(dist)], cache=cache)[0]
        return xloc

    def _mom(self, kloc, cache, **kwargs):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> print(numpy.around(dist.mom([[0, 0, 1], [0, 1, 1]]), 4))
            [1. 0. 0.]
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> print(numpy.around(dist.mom([1, 1]), 4))
            0.5
        """
        if evaluation.get_dependencies(*list(self.inverse_map)):
            raise StochasticallyDependentError(
                "Joint distribution with dependencies not supported.")
        output = 1.
        for dist in evaluation.sorted_dependencies(self):
            if dist not in self.inverse_map:
                continue
            idx = self.inverse_map[dist]
            output *= evaluation.evaluate_moment(
                dist, kloc[idx:idx+len(dist)], cache=cache)
        return output

    def _ttr(self, kloc, cache, **kwargs):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal(), chaospy.Exponential())
            >>> print(numpy.around(dist.ttr([[1, 2, 3], [1, 2, 3], [1, 2, 3]]), 4))
            [[[0.5    0.5    0.5   ]
              [0.     0.     0.    ]
              [3.     5.     7.    ]]
            <BLANKLINE>
             [[0.0833 0.0667 0.0643]
              [1.     2.     3.    ]
              [1.     4.     9.    ]]]
            >>> d0 = chaospy.Uniform()
            >>> dist = chaospy.J(d0, d0+chaospy.Uniform())
            >>> print(numpy.around(dist.ttr([1, 1]), 4)) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            chaospy.distributions.baseclass.StochasticallyDependentError: Joint ...
        """
        if evaluation.get_dependencies(*list(self.inverse_map)):
            raise StochasticallyDependentError(
                "Joint distribution with dependencies not supported.")
        output = numpy.zeros((2,)+kloc.shape)
        for dist in evaluation.sorted_dependencies(self):
            if dist not in self.inverse_map:
                continue
            idx = self.inverse_map[dist]
            values = evaluation.evaluate_recurrence_coefficients(
                dist, kloc[idx:idx+len(dist)], cache=cache)
            output.T[idx] = values.T
        return output


    def __len__(self):
        return sum(len(dist) for dist in self.inverse_map)

    def __str__(self):
        """
        Example:
            >>> print(chaospy.J(chaospy.Uniform(), chaospy.Normal()))
            J(Uniform(lower=0, upper=1), Normal(mu=0, sigma=1))
        """
        args = [str(self.prm[key]) for key in sorted(list(self.prm))]
        return self.__class__.__name__ + "(" + ", ".join(args) + ")"

    def __getitem__(self, i):
        """
        Example:
            >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
            >>> print(dist[0])
            Uniform(lower=0, upper=1)
            >>> print(dist[1])
            Normal(mu=0, sigma=1)
            >>> print(dist[:1])
            J(Uniform(lower=0, upper=1))
            >>> print(dist[:2])
            J(Uniform(lower=0, upper=1), Normal(mu=0, sigma=1))
            >>> print(dist[2])
            Traceback (most recent call last):
                ...
            IndexError: index out of bounds.
        """
        if isinstance(i, int):
            i = "_%03d" % i
            if i in self.prm:
                return self.prm[i]
            raise IndexError("index out of bounds.")
        if isinstance(i, slice):
            start, stop, step = i.start, i.stop, i.step
            if start is None: start = 0
            if stop is None: stop = len(self)
            if step is None: step = 1
            out = []
            prm = self.prm
            for i in range(start, stop, step):
                out.append(prm["_%03d" % i])
            return J(*out)
        raise IndexError("index not recognised.")



import chaospy.poly.collection.arithmetics
import chaospy.poly.collection.core
import chaospy.poly.collection.derivative
import chaospy.poly.collection.numpy_
import chaospy.poly.collection.linalg

from chaospy.poly.collection.arithmetics import *
from chaospy.poly.collection.core import *
from chaospy.poly.collection.derivative import *
from chaospy.poly.collection.numpy_ import *
from chaospy.poly.collection.linalg import *



"""
Polynomial transformation functions involving derivatives.
"""

import numpy as np
from scipy.special import factorial

import chaospy.poly.caller
import chaospy.poly.collection
import chaospy.poly.dimension

from chaospy.poly import Poly


def differential(poly, diffvar):
    """
    Polynomial differential operator.

    Args:
        poly (Poly) : Polynomial to be differentiated.
        diffvar (Poly) : Polynomial to differentiate by. Must be decomposed. If
                polynomial array, the output is the Jacobian matrix.

    Examples:
        >>> q0, q1 = chaospy.variable(2)
        >>> poly = chaospy.Poly([1, q0, q0*q1**2+1])
        >>> print(poly)
        [1, q0, q0q1^2+1]
        >>> print(differential(poly, q0))
        [0, 1, q1^2]
        >>> print(differential(poly, q1))
        [0, 0, 2q0q1]
    """
    poly = Poly(poly)
    diffvar = Poly(diffvar)

    if not chaospy.poly.caller.is_decomposed(diffvar):
        sum(differential(poly, chaospy.poly.caller.decompose(diffvar)))

    if diffvar.shape:
        return Poly([differential(poly, pol) for pol in diffvar])

    if diffvar.dim > poly.dim:
        poly = chaospy.poly.dimension.setdim(poly, diffvar.dim)
    else:
        diffvar = chaospy.poly.dimension.setdim(diffvar, poly.dim)

    qkey = diffvar.keys[0]

    core = {}
    for key in poly.keys:

        newkey = np.array(key) - np.array(qkey)

        if np.any(newkey < 0):
            continue
        newkey = tuple(newkey)
        core[newkey] = poly.A[key] * np.prod(
            [factorial(key[idx], exact=True) / factorial(newkey[idx], exact=True)
             for idx in range(poly.dim)])

    return Poly(core, poly.dim, poly.shape, poly.dtype)


def gradient(poly):
    """
    Gradient of a polynomial.

    Args:
        poly (Poly) : polynomial to take gradient of.

    Returns:
        (Poly) : The resulting gradient.

    Examples:
        >>> q0, q1, q2 = chaospy.variable(3)
        >>> poly = 2*q0 + q1*q2
        >>> print(chaospy.gradient(poly))
        [2, q2, q1]
    """
    return differential(
        poly, chaospy.poly.collection.basis(1, 1, poly.dim, sort="GR"))


def hessian(poly):
    """
    Make Hessian matrix out of a polynomial.

    Args:
        poly (Poly) : polynomial to take Hessian of.

    Returns:
        (Poly) : The resulting Hessian.

    Examples:
        >>> q0, q1, q2 = chaospy.variable(3)
        >>> poly = 2*q0**2 + q1**2*q2
        >>> print(chaospy.hessian(poly))
        [[4, 0, 0], [0, 2q2, 2q1], [0, 2q1, 0]]
    """
    return gradient(gradient(poly))



"""Pareto type 2 distribution."""
from ..baseclass import Dist
from ..operators.addition import Add


class pareto2(Dist):
    """Pareto type 2 distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return c*1.0/(1.0+x)**(c+1.0)

    def _cdf(self, x, c):
        return 1.0-1.0/(1.0+x)**c

    def _ppf(self, q, c):
        return pow(1.0-q,-1.0/c)-1

    def _bnd(self, x, c):
        return 0.0, self._ppf(1-1e-10, c)


class Pareto2(Add):
    """
    Pareto type 2 distribution.

    Also known as Lomax distribution (for loc=0).

    Lower threshold at loc and survival: (1+x)^-shape.

    Args:
        shape (float, Dist): Shape parameter
        scale (float, Dist): Scaling parameter
        loc (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.Pareto2(2, 2, 2)
        >>> print(distribution)
        Pareto2(loc=2, scale=2, shape=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.2361 2.582  3.1623 4.4721]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.7155 0.4648 0.253  0.0894]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.3981 2.126  8.9697 2.7794]
    """

    def __init__(self, shape=1, scale=1, loc=0):
        self._repr = {"shape": shape, "scale": scale, "loc": loc}
        Add.__init__(self, left=pareto2(shape)*scale, right=loc)



import logging
import inspect

import numpy

from ... import quadrature

from .common import DependencyError
from .parameters import load_parameters as load_inputs
from .dependencies import sorted_dependencies, get_dependencies

from .density import evaluate_density
from .forward import evaluate_forward
from .inverse import evaluate_inverse
from .bound import evaluate_bound
from .moment import evaluate_moment
from .recurrence_coefficients import evaluate_recurrence_coefficients


def get_forward_cache(
        distribution,
        cache,
):
    from .. import baseclass
    if not isinstance(distribution, baseclass.Dist):
        return distribution
    if distribution in cache:
        return cache[distribution]
    if hasattr(distribution, "_fwd_cache"):
        return distribution._fwd_cache(cache)
    return distribution

def get_inverse_cache(
        distribution,
        cache,
):
    from .. import baseclass
    if not isinstance(distribution, baseclass.Dist):
        return distribution
    if distribution in cache:
        return cache[distribution]
    if hasattr(distribution, "_inv_cache"):
        return distribution._inv_cache(cache)
    return distribution



"""Folded normal distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class folded_normal(Dist):
    """Folded normal distribution."""

    def __init__(self, c=1):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        return numpy.sqrt(2.0/numpy.pi)*numpy.cosh(c*x)*numpy.exp(-(x*x+c*c)/2.0)

    def _cdf(self, x, c):
        return special.ndtr(x-c) + special.ndtr(x+c) - 1.0

    def _bnd(self, x, c):
        return 0, 7.5+c


class FoldedNormal(Add):
    """
    Folded normal distribution.

    Args:
        mu (float, Dist): Location parameter in normal distribution
        sigma (float, Dist): Scaling parameter (in both normal and fold)
        loc (float, Dist): Location of fold

    Examples:
        >>> distribution = chaospy.FoldedNormal(3, 2, 1)
        >>> print(distribution)
        FoldedNormal(loc=1, mu=3, sigma=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [3.3224 4.4938 5.5067 6.6832]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.1417 0.1934 0.1932 0.14  ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [5.7901 2.6245 8.2952 4.9109]
        >>> print(numpy.around(distribution.mom(1), 4))
        5.034
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[5.3928 6.0914 6.7954]
         [3.7271 6.2602 7.926 ]]
    """

    def __init__(self, mu=0, sigma=1, loc=0):
        self._repr = {"mu": mu, "sigma": sigma, "loc": loc}
        Add.__init__(self, left= folded_normal(mu-loc)*sigma, right=loc)


Foldnormal = deprecation_warning(FoldedNormal, "Foldnormal")



"""Cosines Hyperbolic."""
import numpy

from ..baseclass import Dist
from .. import evaluation, approximation


class Cosh(Dist):
    """
    Cosines Hyperbolic.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Cosh(chaospy.Uniform(0, 1))
        >>> print(distribution)
        Cosh(Uniform(lower=0, upper=1))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [1.0201 1.0811 1.1855 1.3374]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [4.9668 2.4346 1.5707 1.126 ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [1.2213 1.0066 1.4865 1.1185]
        >>> print(numpy.around(distribution.mom(1), 4))
        1.1754
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        """Probability density function."""
        output = evaluation.evaluate_density(dist, numpy.arccosh(x), cache=cache)
        output /= numpy.where(x != 1, numpy.sqrt(x*x-1), numpy.inf)
        return output

    def _cdf(self, x, dist, cache):
        return evaluation.evaluate_forward(dist, numpy.arccosh(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.cosh(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.cosh(evaluation.evaluate_bound(
            dist, numpy.arccosh(x), cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.cosh(dist)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.arccosh(dist)
        return self



"""Anglit distribution."""
import numpy

from ..baseclass import Dist
from ..operators.addition import Add


class anglit(Dist):
    """Anglit distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return numpy.cos(2*x)

    def _cdf(self, x):
        return numpy.sin(x+numpy.pi/4)**2.0

    def _ppf(self, q):
        return (numpy.arcsin(numpy.sqrt(q))-numpy.pi/4)

    def _bnd(self, x):
        return -numpy.pi/4, numpy.pi/4


class Anglit(Add):
    """
    Anglit distribution.

    Args:
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scaling parameter

    Examples:
        >>> distribution = chaospy.Anglit(2, 4)
        >>> print(distribution)
        Anglit(loc=2, scale=4)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [-1.1416  0.9528  2.      3.0472  5.1416]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.     0.2165 0.25   0.2165 0.    ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [2.6245 0.2424 4.2421 1.9288]
        >>> print(numpy.around(distribution.mom([1, 2, 3]), 4))
        [ 2.      5.8696 19.2176]
    """

    def __init__(self, loc=0, scale=1):
        self._repr = {"scale": scale, "loc": loc}
        Add.__init__(self, left=anglit()*scale, right=loc)



"""Polynomial base class Poly."""
import re
import numpy

import chaospy.poly


# Name of variables
VARNAME = "q"
POWER = "^"
SEP = ""


def setvarname(name):
    """Set the variable display name."""
    global VARNAME
    VARNAME = name


class Poly(object):
    """
    Polynomial representation in variable dimensions.

    Can also represent sets of polynomials.

    Examples:
        >>> print(chaospy.Poly({(1,): numpy.array(1)}))
        q0
        >>> x, y = chaospy.variable(2)
        >>> print(x**2 + x*y + 2)
        q0^2+q0q1+2
        >>> poly = -3*x + x**2 + y
        >>> print(poly.coefficients)
        [ 1 -3  1]
        >>> print(poly.exponents)
        [[0 1]
         [1 0]
         [2 0]]
        >>> print(poly([1, 2, 3], [1, 2, 3]))
        [-1  0  3]
        >>> print(chaospy.Poly([x*y, x, y]))
        [q0q1, q0, q1]
    """

    __array_priority__ = 9000

    def __init__(self, core=None, dim=None, shape=None, dtype=None):
        """
        Constructor for the Poly class.

        Args:
            A (numpy.ndarray, dict, Poly) : The polynomial coefficient Tensor.
                    Where A[(i,j,k)] corresponds to a_{ijk} x^i y^j z^k
                    (A[i][j][k] for list and tuple)
            dim (int) : the dimensionality of the polynomial.  Automatically
                    set if A contains a value.
            shape (tuple) : the number of polynomials represented.
                    Automatically set if A contains a value.
            dtype (type) : The type of the polynomial coefficients
        """
        core, dim, shape, dtype = chaospy.poly.constructor.preprocess(
            core, dim, shape, dtype)

        self.keys = sorted(core.keys(), key=sort_key)
        self.dim = dim
        self.shape = shape
        self.dtype = dtype
        self.A = core

    def __abs__(self):
        """Absolute value."""
        core = {key: abs(val) for key, val in self.A.items()}
        return Poly(core, self.dim, self.shape, self.dtype)

    def __add__(self, other):
        """Left addition."""
        return chaospy.poly.collection.arithmetics.add(self, other)

    def __radd__(self, other):
        """Right addition."""
        return chaospy.poly.collection.arithmetics.add(self, other)

    def __mul__(self, other):
        """Left multiplication."""
        return chaospy.poly.collection.arithmetics.mul(self, other)

    def __rmul__(self, other):
        """Right multiplication."""
        return chaospy.poly.collection.arithmetics.mul(self, other)

    def __sub__(self, other):
        """Left subtraction."""
        return chaospy.poly.collection.arithmetics.add(self, -other)

    def __rsub__(self, other):
        """Right subtraction."""
        return chaospy.poly.collection.arithmetics.add(-self, other)

    def __div__(self, other):
        """Python2 division."""
        return chaospy.poly.collection.arithmetics.mul(
            self, numpy.asfarray(other)**-1)
        return NotImplemented

    def __truediv__(self, other):
        """True division."""
        return self.__div__(other)

    def __floordiv__(self, other):
        return chaospy.poly.collection.asint(self.__div__(other))

    def __eq__(self, other):
        return ~self.__ne__(other)

    def __ne__(self, other):
        if not isinstance(other, Poly):
            other = Poly(other)
        diff = abs(self - other)
        out = numpy.zeros(diff.shape, dtype=bool)
        for key in diff.keys:
            out = out + (diff.A[key]!=0)
        return out

    def __call__(self, *args, **kws):
        """
        Evaluate a polynomial.

        Args:
            args (numpy.ndarray, Poly) : Arguments to evaluate. Masked values and
                    numpy.nan will not be evaluated. If instance is Poly,
                    substitution on the variable is performed.
            kws (numpy.ndarray, Poly) : Same as args, but the keys referred to the
                    variables names. If the number of dimensions are <=3, 'x',
                    'y' and 'z' respectably refer to the axes. Otherwise, the
                    keys are on the form 'x%d' where %d is an interger
                    representing the dimension.

        Returns:
            (ndarray, Poly) : If masked values are included in args, a Poly is
                    returned where the masked variables are retained.
                    Otherwise an array is returned.
        """
        if len(args)>self.dim:
            args = list(args[:self.dim])
        else:
            args = list(args) + [numpy.nan]*(self.dim-len(args))

        for key,val in kws.items():

            if key[0]=="q":
                index = int(key[1:])

                if index>=self.dim:
                    continue

            else:
                raise TypeError(
                    "Unexpeted keyword argument '%s'" % key)

            if args[index] not in (numpy.nan, numpy.ma.masked):
                raise TypeError(
                    "Multiple values for keyword argument '%s'" % index)

            args[index] = val

        return chaospy.poly.caller.call(self, args)


    def __contains__(self, y):
        """x.__contains__(y) <==> y in x"""

        if not isinstance(y, Poly):
            y = Poly(y)

        if self.shape==():
            if len(y.keys)>1:
                return NotImplemented
            return y.keys[0] in self.keys

        if len(y.shape)==len(self.shape)-1:
            return max(map(y.__eq__, self))

        if len(y.shape)<len(self.shape)-1:
            return max([y in s for s in self])

        return NotImplemented


    def __getitem__(self, I):
        shape = self.A[self.keys[0]][I].shape

        if isinstance(I, numpy.ndarray):
            I = I.tolist()

        if isinstance(I, (slice, int)):
            I = (I,)

        if not self.shape:
            subkey = I[0]
            subset = slice(None, None, None)

        elif len(self.shape)>=len(I):
            subkey = slice(None, None, None)
            subset = I

        elif len(self.shape)+1==len(I):
            subkey = I[-1]
            subset = I[:-1]

        else:
            subkey = slice(None, None, None)
            subset = I

        if isinstance(subkey, int):
            if not 0<=subkey<len(self.keys):
                raise IndexError("Index out of range")

            subkey = slice(subkey, subkey+1,None)


        A0 = {}
        for key in self.keys:
            tmp = self.A[key][subset]
            if not numpy.all(tmp==0):
                A0[key] = tmp

        A1 = {}
        for key in list(A0.keys())[subkey]:
            A1[key] = A0[key]

        return Poly(A1, self.dim, shape, self.dtype)

    def __iter__(self):
        """x.__iter__() <==> iter(x)"""
        A = self.A
        Pset = []

        if not self.shape:
            raise ValueError("not iterable")

        for i in range(self.shape[0]):

            out = {}
            for key in self.keys:

                if numpy.any(A[key][i]):
                    out[key] = A[key][i]

            Pset.append(Poly(out, self.dim, self.shape[1:],self.dtype))

        return Pset.__iter__()

    def __len__(self):
        """x.__len__() <==> len(x)"""

        if self.shape:
            return self.shape[0]
        return 1

    def __neg__(self):
        """x.__neg__() <==> -x"""

        A = self.A.copy()
        for key in self.keys:
            A[key] = -A[key]
        return Poly(A, self.dim, self.shape, self.dtype)

    def __pos__(self):
        return Poly(self.A, self.dim, self.shape, self.dtype)

    def __nonzero__(self):
        """x.__nonzero__() <==> x != 0"""

        if not self.A:
            return False
        return True


    def __pow__(self, n):
        """x.__pow__(y) <==> x**y"""
        if isinstance(n, (int, float, numpy.generic)):
            n = numpy.array(n)
            assert isinstance(n, numpy.ndarray)

        if isinstance(n, numpy.ndarray) and not n.shape:

            if abs(n-int(n))>1e-5:
                raise ValueError("Power of Poly must be interger")
            n = int(n)

            if n == 0:
                return Poly(
                    {(0,)*self.dim: numpy.ones(self.shape, dtype=int)},
                    self.dim, self.shape, None)

            self = self.copy()
            out = self
            for poly in range(1, n):
                out = out * self
            return out

        elif isinstance(n, (numpy.ndarray, list, tuple)):

            if not self.shape:
                out = [self.__pow__(n[i]) for i in range(len(n))]
                return Poly(out, self.dim, None, None)

            return Poly([self[i].__pow__(n[i])
                for i in range(len(n))], self.dim, None, None)

        raise ValueError("unknown type %s (%s)" % (str(n), type(n)))
        return NotImplemented



    def __str__(self):
        """x.__str__() <==> str(x)"""
        self = self.copy()

        # Array
        if self.shape:

            if len(self.shape) > 1:
                shape = self.shape
                self = chaospy.poly.shaping.flatten(self)
                P = numpy.reshape(numpy.array([str(p) for p in self],\
                    dtype=object), shape)
            else:
                P = numpy.array([str(p) for p in self], dtype=object)
            out = str(P.tolist())
            out = "".join(out.split("'"))
            return out

        if isinstance(VARNAME, str):
            basename = ["%s%d" % (VARNAME, d) for d in range(self.dim)]
        else:
            basename = list(VARNAME)

        out = []
        for key in self.keys[::-1]:

            o = ""
            coef = self.A[key]
            if numpy.sign(coef) == 1:
                o += "+"
            else:
                o += "-"
            if isinstance(coef, numpy.ndarray):
                o += str(abs(coef.item()))
            else:
                o += str(abs(coef))

            for j in range(self.dim):

                if key[j] == 1:
                    o += basename[j]
                if key[j] > 1:
                    o +="%s%s%d" % (basename[j], POWER, key[j])

            if not o:
                if self.dtype in (int, numpy.int16, numpy.int32, numpy.int64):
                    out = ["0"]
                else:
                    out = ["0.0"]
            elif o == "-0":
                o = "+0"
            if o == "-0.0":
                o = "+0.0"

            for b in basename:
                if o[4:5] == b[0] and o[1:].startswith("1.0"):
                    o = o[0] + o[4:]
                elif o[2:3] == b[0] and o[1:2] == "1":
                    o = o[0] + o[2:]
                elif ("/1%s" % b[0]) in o:
                    o = b.join(o.split("/1%s" % b))

            out.append(o)

        if not out:
            if self.dtype in (int, numpy.int16, numpy.int32, numpy.int64):
                out = ["0"]
            else:
                out = ["0.0"]
        if out[0][0]=="+": out[0] = out[0][1:]

        out = "".join(out)
        out = re.sub(r"(\d)(%s)" % re.escape(VARNAME), r"\1%s\2" % SEP, out)
        return out


    def copy(self):
        """Return a copy of the polynomial."""
        return Poly(self.A.copy(), self.dim, self.shape,
            self.dtype)

    @property
    def coefficients(self):
        """Polynomial coefficients."""
        out = numpy.array([self.A[key] for key in self.keys])
        out = numpy.rollaxis(out, -1)
        return out

    @property
    def exponents(self):
        """Polynomial exponents."""
        out = numpy.array(self.keys)
        return out

def sort_key(val):
    """Sort key for sorting keys in grevlex order."""
    maxval = max(val)
    lenval = len(val)
    exponents = range(lenval-1, -1, -1)
    r = [(maxval+1)**expo*val[lenval-1-expo] for expo in exponents]
    return sum(r)



"""Power log-Normal probability distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class power_log_normal(Dist):
    """Power log-Normal probability distribution."""

    def __init__(self, c, s):
        Dist.__init__(self, c=c, s=s)

    def _pdf(self, x, c, s):
        norm = (2*numpy.pi)**-.5*numpy.exp(-(numpy.log(x)/s)**2/2.)
        return c/(x*s)*norm*pow(special.ndtr(-numpy.log(x)/s), c*1.-1.)

    def _cdf(self, x, c, s):
        return 1. - pow(special.ndtr(-numpy.log(x)/s), c*1.)

    def _ppf(self, q, c, s):
        return numpy.exp(-s*special.ndtri(pow(1.-q, 1./c)))

    def _bnd(self, x, c, s):
        return 0., self._ppf(1-1e-10, c, s)


class PowerLogNormal(Add):
    """
    Power log-normal distribution

    Args:
        shape (float, Dist) : Shape parameter
        mu (float, Dist) : Mean in the normal distribution.  Overlaps with
                scale by mu=log(scale)
        sigma (float, Dist) : Standard deviation of the normal distribution.
        shift (float, Dist) : Location parameter
        scale (float, Dist) : Scaling parameter. Overlap with mu in scale=e**mu

    Examples:
        >>> distribution = chaospy.PowerLogNormal(2, 2, 2, 2, 2)
        >>> print(distribution)
        PowerLogNormal(mu=2, scale=2, shape=2, shift=2, sigma=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [ 3.212   5.2707  9.5114 21.2701]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.1347 0.0711 0.0317 0.0092]
        >>> print(numpy.around(distribution.sample(4), 4))
        [11.4445  2.6512 69.8654  6.6177]
    """
    def __init__(self, shape=1, mu=0, sigma=1, shift=0, scale=1):
        self._repr = {
            "shape": shape, "mu": mu, "sigma": sigma, "shift": shift, "scale": scale}
        Add.__init__(self, left=power_log_normal(
            shape, sigma)*scale*numpy.e**mu, right=shift)


PowerLognormal = deprecation_warning(PowerLogNormal, "PowerLognormal")



"""Gamma distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class gamma(Dist):

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        return x**(a-1)*numpy.e**(-x) / special.gamma(a)

    def _cdf(self, x, a):
        return special.gammainc(a, x)

    def _ppf(self, q, a):
        return special.gammaincinv(a, q)

    def _mom(self, k, a):
        return special.gamma(a+k)/special.gamma(a)

    def _ttr(self, n, a):
        return 2.*n+a, n*n+n*(a-1)

    def _bnd(self, x, a):
        return 0, 40+2*a


class Gamma(Add):
    """
    Gamma distribution.

    Also an Erlang distribution when shape=k and scale=1./lamb.

    Args:
        shape (float, Dist):
            Shape parameter. a>0.
        scale (float, Dist):
            Scale parameter. scale!=0
        shift (float, Dist):
            Location of the lower bound.

    Examples:
        >>> distribution = chaospy.Gamma(1, 1, 1)
        >>> print(distribution)
        Gamma(scale=1, shape=1, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [1.2231 1.5108 1.9163 2.6094]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.8 0.6 0.4 0.2]
        >>> print(numpy.around(distribution.sample(4), 4))
        [2.0601 1.1222 4.0014 1.6581]
        >>> print(distribution.mom(1))
        2.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[4. 6. 8.]
         [1. 4. 9.]]
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=gamma(shape)*scale, right=shift)


class Exponential(Add):
    R"""
    Exponential Probability Distribution

    Args:
        scale (float, Dist) : Scale parameter. scale!=0
        shift (float, Dist) : Location of the lower bound.

    Examples;:
        >>> distribution = chaospy.Exponential(2, 3)
        >>> print(distribution)
        Exponential(scale=2, shift=3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [3.4463 4.0217 4.8326 6.2189]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.sample(4), 4))
        [5.1203 3.2444 9.0028 4.3163]
        >>> print(numpy.around(distribution.mom(1), 4))
        5.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 9. 13. 17.]
         [ 4. 16. 36.]]
    """

    def __init__(self, scale=1, shift=0):
        self._repr = {"scale": scale, "shift": shift}
        Add.__init__(self, left=gamma(1)*scale, right=shift)



"""
Implementation of Schnabel-Eskow modified cholesky factorisation.

Schnabel, R. B. and Eskow, E. 1999, mat revised modifed cholesky factorisation
algorithm. SIAM J. Optim. 9, 1135-1148.
"""
import numpy


def schnabel_eskow(mat, eps=1e-16):
    """
    Scnabel-Eskow algorithm for modified Cholesky factorisation algorithm.

    Args:
        mat (numpy.ndarray) : Must be a non-singular and symmetric matrix If
            sparse, the result will also be sparse.
        eps (float) : Error tolerance used in algorithm.

    Returns
    -------
    perm : 2d array
        Permutation matrix used for pivoting.
    lowtri : 2d array
        Lower triangular factor
    err : 1d array
        Positive diagonals of shift matrix `err`.

    Examples
    --------
    >>> mat = [[4, 2, 1], [2, 6, 3], [1, 3, -.004]]
    >>> perm, lowtri, err = schnabel_eskow(mat)
    >>> perm, lowtri = numpy.matrix(perm), numpy.matrix(lowtri)
    >>> print(numpy.around(perm*lowtri*lowtri.T*perm.T, 4))
    [[5.504 2.    1.   ]
     [2.    6.    3.   ]
     [1.    3.    1.5  ]]
    """
    mat = numpy.asfarray(mat)
    tau = eps**(1/3.)

    # Create the matrix err and mat.
    size = mat.shape[0]
    mat0 = mat
    mat = 1*mat
    err = numpy.zeros(size, dtype=float)

    # Permutation matrix.
    perm = numpy.eye(size)

    # Calculate gamma.
    gamma = abs(mat.diagonal()).max()

    # Phase one, mat potentially positive definite.
    ###############################################

    def invariant(mat, k):
        """Return `True` if the invariant is satisfied."""
        A_ = numpy.eye(size)
        L_ = numpy.eye(size)
        A_[k:, k:] = numpy.triu(mat[k:, k:], 0) + numpy.triu(mat[k:, k:], 1).T
        L_[:, :k] = numpy.tril(mat[:, :k])
        return numpy.allclose(
            numpy.dot(L_, numpy.dot(A_, L_.T)),
            numpy.dot(perm, numpy.dot(mat0, perm.T)) + numpy.diag(err)
        )

    def jiter_factor(mat, j, perm, err):
        """Perform jth iteration of factorisation.
        """
        assert invariant(mat, j)

        mat[j, j] = numpy.sqrt(mat[j, j])
        mat[j+1:, j] /= mat[j, j]
        mat[j+1:, j+1:] -= mat[j+1:, j:j+1]*mat[j+1:, j:j+1].T
        mat[j, j+1:] = 0

        assert invariant(mat, j+1)

    def permute(mat, perm, i, j):
        """Exchange rows and columns i and j of mat and recored the
        permutation in perm"""
        p = numpy.arange(size, dtype=int)
        if i != j:
            p[[i, j]] = j, i

            perm[::] = perm[p, :]
            mat[::] = mat[p, :]
            mat[::] = mat[:, p]

    def exec_phasetwo(mat, perm, err, j):
        """Phase 2 of the algorithm, mat not positive definite."""
        if j == size:
            # delta = err[size].
            delta = -mat[size-1, size-1] + max(
                -tau*mat[size-1, size-1]/(1 - tau), tau*tau*gamma)

            err[size-1] = delta
            mat[size-1, size-1] += delta
            mat[size-1, size-1] = numpy.sqrt(mat[size-1, size-1])

        else:
            # Number of iterations performed in phase one (less 1).
            k = j - 1

            # Calculate the lower Gerschgorin bounds of Ak+1.
            tmp = mat[k+1:, k+1:]
            g = tmp.diagonal().copy()
            tmp = abs(numpy.tril(tmp, -1))
            g -= tmp.sum(axis=0)
            g -= tmp.sum(axis=1)

            # Modified Cholesky decomposition.
            delta_prev = 0.0
            for j in range(k+1, size-2):
                # Pivot on the maximum lower Gerschgorin bound
                # estimate.
                i = j + numpy.argmax(g[j-(k+1):])

                # Interchange row and column i and j.
                permute(mat, perm, i, j)

                # Calculate err[j] and add to diagonal.
                normj = abs(mat[j+1:, j]).sum()
                delta = max(0.0,
                            -mat[j, j] + max(normj, tau*tau*gamma),
                            delta_prev)  # delta = E[size].

                if delta > 0.0:
                    mat[j, j] += delta
                    delta_prev = delta
                err[j] = delta

                # Update Gerschgorin bound estimates.
                if mat[j, j] != normj:
                    temp = 1.0 - normj / mat[j, j]
                    g[j-k:] += abs(mat[j+1:, j])*temp

                # Perform jth iteration of factorisation.
                jiter_factor(mat, j, perm, err)

            # Final 2*2 submatrix.
            mini = mat[-2:, -2:]
            mini[1, 0] = mini[0, 1]
            eigs = numpy.sort(numpy.linalg.eigvalsh(mini))
            delta = max(
                0, -eigs[0] + max(tau*(eigs[1] - eigs[0])/(1 - tau),
                                  tau*tau*gamma), delta_prev)
            if delta > 0.0:
                mat[size-2, size-2] += delta
                mat[size-1, size-1] += delta
                delta_prev = delta
            err[size-2] = err[size-1] = delta

            mat[size-2, size-2] = numpy.sqrt(mat[size-2, size-2])
            mat[size-1, size-2] /= mat[size-2, size-2]
            mat[size-1, size-1] = numpy.sqrt(
                mat[size-1, size-1] - mat[size-1, size-2]**2)

    for j in range(size):

        # Calculate max_Aii and min_Aii
        diag = mat.diagonal()[j:]

        # Test for phase 2, mat not positive definite.
        if diag.max() < tau*tau * gamma or diag.min() < - 0.1 * diag.max():
            exec_phasetwo(mat, perm, err, j)
            break
        else:
            # Pivot on maximum diagonal of remaining submatrix.
            i = j + numpy.argmax(mat.diagonal()[j:])

            # Interchange row and column i and j.
            permute(mat, perm, i, j)

            # Test for phase 2 again.
            min_num = 1e99
            mat_diag = mat.diagonal()
            if j + 1 < size:
                min_num = (mat_diag[i] -
                           mat[i, j+1:]**2/mat_diag[j+1:]).min()
            else:
                min_num = mat_diag[i]

            if j+1 <= size and min_num < - 0.1 * gamma:
                exec_phasetwo(mat, perm, err, j)
                break

            # Perform jth iteration of factorisation.
            else:
                jiter_factor(mat, j, perm, err)

    # The Cholesky factor of mat.
    return perm, numpy.tril(mat), err



r"""
One of the basic assumption in polynomial chaos expansion is that a collection
of polynomials that are mutually orthogonal on a weighted Hilbert space. The
relation is typically written mathematically as:

.. math::
    \left\langle \Phi_n, \Phi_m \right\rangle = 0 \qquad n \neq m

In practice this relation is instead expressed by the equivalent notation using
expected values:

.. math::
    \mathbb E[\Phi_n \Phi_m] = 0 \qquad n \neq m

This relationship can in the case of ``chaospy`` be expressed as follows::

    >>> polynomials = chaospy.prange(4)
    >>> print(polynomials)
    [1, q0, q0^2, q0^3]
    >>> distribution = chaospy.Normal(0, 1)
    >>> print(chaospy.E(polynomials[0]*polynomials[1], distribution))
    0.0

Note that here the expected value operator implicitly assumes a probability
space to take expected value over, which in the case of the software have to be
defined explicitly.

For a more convenient way to check all pairs at the same time, it is possible
to use the ``chaospy.outer`` function::

    >>> print(chaospy.E(chaospy.outer(polynomials, polynomials), distribution))
    [[ 1.  0.  1.  0.]
     [ 0.  1.  0.  3.]
     [ 1.  0.  3.  0.]
     [ 0.  3.  0. 15.]]

For the case of ``chaospy.prange`` (which creates simple monomials), not all
pairs are orthogonal, as the outer matrix is not diagonal. The methods for
creating orthogonal polynomials are listed bellow.
"""
from .three_terms_recursion import orth_ttr
from .lagrange import lagrange_polynomial
from .gram_schmidt import orth_gs
from .cholesky import orth_chol



"""Levy distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class levy(Dist):
    """Levy distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return 1/numpy.sqrt(2*numpy.pi*x)/x*numpy.exp(-1/(2*x))

    def _cdf(self, x):
        return 2*(1-special.ndtr(1/numpy.sqrt(x)))

    def _ppf(self, q):
        val = special.ndtri(1-q/2.0)
        return 1.0/(val*val)

    def _bnd(self, x):
        return 0., self._ppf(1-1e-10)


class Levy(Add):
    """
    Levy distribution

    Args:
        loc (float, Dist): Location parameter
        scale (float, Dist): Scaling parameter

    Examples:
        >>> distribution = chaospy.Levy(2, 2)
        >>> print(distribution)
        Levy(loc=2, scale=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [ 3.2177  4.8236  9.2728 33.16  ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.1847 0.0834 0.0251 0.0031]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 11.9303   2.8051 516.4406   6.0494]
    """
    def __init__(self, loc=0, scale=1):
        self._repr = {"loc": loc, "scale": scale}
        Add.__init__(self, left=levy()*scale, right=loc)



"""
Evaluate three terms recurrence coefficients (TTR).

Example usage
-------------

Define a simple distribution and data::

    >>> class Exponential(chaospy.Dist):
    ...     def _ttr(self, k_data, alpha): return (2*k_data+1)/alpha, (k_data/alpha)**2
    >>> dist = Exponential(alpha=2.)
    >>> k_data = numpy.array([2])

Normal usage::

    >>> print(evaluate_recurrence_coefficients(dist, k_data))
    [2.5 1. ]
    >>> print(evaluate_recurrence_coefficients(dist, k_data, parameters={"alpha": 1.}))
    [5. 4.]

The use of cache::

    >>> print(evaluate_recurrence_coefficients(dist, k_data, cache={((2,), dist): (3., 4.)}))
    (3.0, 4.0)

Approximate with the use of density, forward, inverse and bound function if recurrence function is missing::

    >>> class Exponential(chaospy.Dist):
    ...     def _pdf(self, x_data, alpha): return alpha*numpy.e**(-alpha*x_data)
    ...     def _cdf(self, x_data, alpha): return 1-numpy.e**(-alpha*x_data)
    ...     def _ppf(self, u_data, alpha): return -numpy.log(1-u_data)/alpha
    ...     def _bnd(self, x_data, alpha): return 0, 20
    >>> dist = Exponential(alpha=2)
    >>> print(evaluate_recurrence_coefficients(dist, k_data))
    [2.5 1. ]
"""
import logging

import numpy
from .parameters import load_parameters


def evaluate_recurrence_coefficients(
        distribution,
        k_data,
        parameters=None,
        cache=None,
):
    """
    Evaluate three terms recurrence coefficients (TTR).

    Args:
        distribution (Dist):
            Distribution to evaluate.
        x_data (numpy.ndarray):
            Locations for where evaluate recurrence coefficients for.
        parameters (:py:data:typing.Any):
            Collection of parameters to override the default ones in the
            distribution.
        cache (:py:data:typing.Any):
            A collection of previous calculations in case the same distribution
            turns up on more than one occasion.

    Returns:
        The recurrence coefficients ``A`` and ``B`` of ``distribution`` at
        location ``x_data`` using parameters ``parameters``.
    """
    assert len(k_data) == len(distribution), (
        "distribution %s is not of length %d" % (distribution, len(k_data)))
    assert len(k_data.shape) == 1

    def cache_key(distribution):
        return (tuple(k_data), distribution)

    if cache is None:
        cache = {}
    else:
        if cache_key(distribution) in cache:
            return cache[cache_key(distribution)]

    try:
        parameters = load_parameters(
            distribution, "_ttr", parameters, cache, cache_key)
        coeff1, coeff2 = distribution._ttr(k_data, **parameters)

    except NotImplementedError:
        from ...quadrature import generate_quadrature
        from ...quadrature.recurrence.stieltjes import discretized_stieltjes
        abscissas, weights = generate_quadrature(100, distribution, rule="C")
        (coeff1, coeff2), _, _ = discretized_stieltjes(
            numpy.max(k_data), abscissas, weights, normed=False)
        range_ = numpy.arange(len(distribution), dtype=int)
        coeff1 = coeff1[range_, k_data]
        coeff2 = coeff2[range_, k_data]

    out = numpy.zeros((2,) + k_data.shape)
    out.T[:, 0] = numpy.asarray(coeff1).T
    out.T[:, 1] = numpy.asarray(coeff2).T
    if len(distribution) == 1:
        out = out[:, 0]

    return out



"""
Functions related to the type of the coefficients.
"""
from __future__ import division

import numpy

from chaospy.poly.base import Poly

DATATYPES = [
    Poly,
    float,
    numpy.float16,
    numpy.float32,
    numpy.float64,
    int,
    numpy.int16,
    numpy.int32,
    numpy.int64,
]

def dtyping(*args):
    """
    Find least common denominator dtype.

    Examples:
        >>> str(dtyping(int, float)) in ("<class 'float'>", "<type 'float'>")
        True
        >>> print(dtyping(int, Poly))
        <class 'chaospy.poly.base.Poly'>
    """
    args = list(args)

    for idx, arg in enumerate(args):

        if isinstance(arg, Poly):
            args[idx] = Poly

        elif isinstance(arg, numpy.generic):
            args[idx] = numpy.asarray(arg).dtype

        elif isinstance(arg, (float, int)):
            args[idx] = type(arg)

    for type_ in DATATYPES:
        if type_ in args:
            return type_

    raise ValueError(
        "dtypes not recognised " + str([str(_) for _ in args]))


def asfloat(vari, limit=None):
    """
    Convert dtype of polynomial coefficients to float.

    Example:
        >>> poly = 2*cp.variable()+1
        >>> print(poly)
        2q0+1
        >>> print(cp.asfloat(poly))
        2.0q0+1.0
    """
    if limit is None:
        limit = 10**300
    if isinstance(vari, Poly):
        core = vari.A.copy()
        for key in vari.keys:
            core[key] = core[key]*1.
        return Poly(core, vari.dim, vari.shape, float)

    return numpy.asfarray(vari)


def asint(vari):
    """
    Convert dtype of polynomial coefficients to float.

    Example:
        >>> poly = 1.5*cp.variable()+2.25
        >>> print(poly)
        1.5q0+2.25
        >>> print(cp.asint(poly))
        q0+2
    """
    if isinstance(vari, Poly):

        core = vari.A.copy()
        for key in vari.keys:
            core[key] = numpy.asarray(core[key], dtype=int)

        return Poly(core, vari.dim, vari.shape, int)

    return numpy.asarray(vari, dtype=int)



def tolist(poly):
    """
    Convert polynomial array into a list of polynomials.

    Examples:
        >>> poly = cp.prange(3)
        >>> print(poly)
        [1, q0, q0^2]
        >>> array = cp.tolist(poly)
        >>> print(isinstance(array, list))
        True
        >>> print(array[1])
        q0
    """
    return toarray(poly).tolist()


def toarray(vari):
    """
    Convert polynomial array into a numpy.asarray of polynomials.

    Args:
        vari (Poly, numpy.ndarray):
            Input data.

    Returns:
        (numpy.ndarray):
            A numpy array with ``Q.shape==A.shape``.

    Examples:
        >>> poly = cp.prange(3)
        >>> print(poly)
        [1, q0, q0^2]
        >>> array = cp.toarray(poly)
        >>> print(isinstance(array, numpy.ndarray))
        True
        >>> print(array[1])
        q0
    """
    if isinstance(vari, Poly):
        shape = vari.shape
        out = numpy.asarray(
            [{} for _ in range(numpy.prod(shape))],
            dtype=object
        )
        core = vari.A.copy()
        for key in core.keys():

            core[key] = core[key].flatten()

            for i in range(numpy.prod(shape)):

                if not numpy.all(core[key][i] == 0):
                    out[i][key] = core[key][i]

        for i in range(numpy.prod(shape)):
            out[i] = Poly(out[i], vari.dim, (), vari.dtype)

        out = out.reshape(shape)
        return out

    return numpy.asarray(vari)


import chaospy as cp # pylint: disable=unused-import

if __name__ == '__main__':
    import doctest
    doctest.testmod()



r"""
Quadrature methods, or numerical integration, is broad class of algorithm for
performing integration of any function :math:`g` that are defined without
requiring an analytical definition. In the scope of ``chaospy`` we limit this
scope to focus on methods that can be reduced to the following approximation:

.. math::
    \int p(x) g(x) dx \approx \sum_{n=1}^N W_n g(X_n)

Here :math:`p(x)` is an weight function, which is assumed to be an probability
distribution, and :math:`W_n` and :math:`X_n` are respectively quadrature
weights and abscissas used to define the approximation.

The simplest example applying such an approximation is Monte Carlo integration.
In such a method, you only need to select :math:`W_n=1/N` and :math:`X_n` to be
independent identical distributed samples drawn from the distribution of
:math:`p(x)`. In practice::

    >>> distribution = chaospy.Uniform(-1, 1)
    >>> abscissas = distribution.sample(1000)
    >>> weights = 1./len(abscissas)

However, except for very high dimensional problems, Monte Carlo is quite an
inefficient way to perform numerical integration, and there exist quite a few
methods that performs better in most low-dimensional settings. If however,
Monto Carlo is your best choice, it might be worth taking a look at
:ref:`sampling`.

.. note:
    Most quadrature rules optimized to a given weight function is referred to
    as the :ref:`gaussian` rules. It does the embedding of the weight function
    automatically as that is what it is designed for. For most other quadrature
    rules, including a weight function is typically not canonical. This however
    isn't very compatible with the Gaussian quadrature rules which take
    probability density functions into account as part of their implementation.
    It also does not match well with ``chaospy`` which assumes the density
    weight function to be defined and incorporated implicit.

    To address this issue, the weight functions are incorporated into the
    weight terms by substituting :math:`W^*_i \leftarrao W_i p(X_i)`, giving
    us:

    .. math::
        \int p(x) g(x) dx \approx
        \sum_i W_i p(X_i) g(X_i) = \sum_i W^{*}_i g(X_i)

    Which is the same format as the Gaussian quadrature rules.

    The consequence of this is that non-Gaussian quadrature rules only produces
    the canonical weights for the probability distribution
    ``chaospy.Uniform(0, 1)``, everything else is custom. To get around this
    limitation, there are few workarounds:

    * Use a uniform distribution on an arbitrary interval ``Uniform(a, b)``,
      and multiply the weight terms with the interval length: ``W *= (b-a)``
    * Use the quadrature rules directly from ``chaospy.quadrature.collection``.
    * Adjust weights afterwards: ``W /= dist.pdf(X)``

To create quadrature abscissas and weights, use the
:func:`~chaospy.quadrature.frontend.generate_quadrature` function. Which type
of quadrature to use is defined by the flag ``rule``. This argument can either
be the full name, or a single letter representing the rule. These are as
follows.

"""
from .frontend import generate_quadrature
from .sparse_grid import construct_sparse_grid
from .combine import combine

from .clenshaw_curtis import quad_clenshaw_curtis
from .fejer import quad_fejer
from .gaussian import quad_gaussian
from .gauss_patterson import quad_gauss_patterson
from .gauss_legendre import quad_gauss_legendre
from .gauss_lobatto import quad_gauss_lobatto
from .gauss_kronrod import quad_gauss_kronrod
from .gauss_radau import quad_gauss_radau
from .genz_keister import quad_genz_keister
from .leja import quad_leja
from .newton_cotes import quad_newton_cotes

from .recurrence import (
    analytical_stieljes,
    modified_chebyshev,
    coefficients_to_quadrature,
    construct_recurrence_coefficients,
    discretized_stieltjes,
    lanczos,
    RECURRENCE_ALGORITHMS,
)



"""
Parameter loading.

Example usage
-------------

Define a simple distribution::

    >>> class MyDist(chaospy.Dist):
    ...     def _pdf(self, x_data, param1): pass
    >>> dist = MyDist(param1=45)

Normal usage::

    >>> print(load_parameters(dist, "_pdf"))
    {'param1': 45}
    >>> print(load_parameters(dist, "_pdf", parameters={"param1": 35}))
    {'param1': 35}

Define distribution dependencies::

    >>> sub_dist = chaospy.Uniform(0, 1)
    >>> dist = MyDist(param1=sub_dist)

Unresolved usage::

    >>> load_parameters(dist, "_pdf") # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    chaospy.distributions.baseclass.StochasticallyDependentError: \
evaluating under-defined distribution MyDist(param1=Uniform(lower=0, upper=1)).

Loading combined with caching::

    >>> print(load_parameters(dist, "_pdf", cache={sub_dist: 15}))
    {'param1': 15}

Advanced distributions which has the ``cache`` as a call argument, receive the
cache as the value::

    >>> class MyAdvancedDist(chaospy.Dist):
    ...     def _pdf(self, x_data, param1, cache): pass
    >>> dist = MyAdvancedDist(param1=sub_dist)
    >>> params = load_parameters(dist, "_pdf", cache={sub_dist: 15})
    >>> print(sorted(params))
    ['cache', 'param1']
    >>> print(params["param1"])
    Uniform(lower=0, upper=1)
    >>> print(params["cache"])
    {Uniform(lower=0, upper=1): 15}
"""
import inspect

from .common import contains_call_signature


def load_parameters(
        distribution,
        method_name,
        parameters=None,
        cache=None,
        cache_key=lambda x:x,
):
    """
    Load parameter values by filling them in from cache.

    Args:
        distribution (Dist):
            The distribution to load parameters from.
        method_name (str):
            Name of the method for where the parameters should be used.
            Typically ``"_pdf"``, ``_cdf`` or the like.
        parameters (:py:data:typing.Any):
            Default parameters to use if there are no cache to retrieve. Use
            the distributions internal parameters, if not provided.
        cache (:py:data:typing.Any):
            A dictionary containing previous evaluations from the stack. If
            a parameters contains a distribution that contains in the cache, it
            will be replaced with the cache value. If omitted, a new one will
            be created.
        cache_key (:py:data:typing.Any)
            Redefine the keys of the cache to suite other purposes.

    Returns:
        Same as ``parameters``, if provided. The ``distribution`` parameter if
        not. In either case, parameters may be updated with cache values (if
        provided) or by ``cache`` if the call signature of ``method_name`` (on
        ``distribution``) contains an ``cache`` argument.
    """
    from .. import baseclass
    if cache is None:
        cache = {}
    if parameters is None:
        parameters = {}
    parameters_ = distribution.prm.copy()
    parameters_.update(**parameters)
    parameters = parameters_

    # self aware and should handle things itself:
    if contains_call_signature(getattr(distribution, method_name), "cache"):
        parameters["cache"] = cache

    # dumb distribution and just wants to evaluate:
    else:
        for key, value in parameters.items():
            if isinstance(value, baseclass.Dist):
                value = cache_key(value)
                if value in cache:
                    parameters[key] = cache[value]
                else:
                    raise baseclass.StochasticallyDependentError(
                        "evaluating under-defined distribution {}.".format(distribution))

    return parameters



"""Discretized Stieltjes' method."""
import numpy
import chaospy.poly


def discretized_stieltjes(order, abscissas, weights, normed=False):
    """
    Discretized Stieltjes' method.

    Args:
        order (int):
            The polynomial order create.
        abscissas (numpy.ndarray):
            Quadrature abscissas, assumed to be of ``shape == (D, N)``, where
            ``D`` is the number of distributions to handle at once, and ``N``
            is the number of abscissas.
        weights (numpy.ndarray):
            Quadrature weights, assumed to be of ``shape == (N,)``.
        normed (bool):
            If true, normalize polynomials.

    Returns:
        (numpy.ndarray, numpy.ndarray, numpy.ndarray):
            coefficients:
                The recurrence coefficients created using the discretized
                Stieltjes' method, with ``shape == (2, D, order+1)``.
            polynomials:
                The orthogonal polynomial expansion created as a by-product of
                the algorithm.
            norms:
                The norm of each orthogonal polynomial. Roughly equivalent to
                ``chaospy.E(polynomials**2, dist)``, but more numerically
                stable than most alternatives.

    Examples:
        >>> dist = chaospy.J(chaospy.Uniform(0, 1), chaospy.Beta(3, 4))
        >>> abscissas, weights = chaospy.generate_quadrature(
        ...     9, dist, rule="clenshaw_curtis")
        >>> coeffs, orth, norms = discretized_stieltjes(2, abscissas, weights)
        >>> print(numpy.around(coeffs, 5))
        [[[0.5     0.5     0.5    ]
          [0.42857 0.46032 0.47525]]
        <BLANKLINE>
         [[1.      0.08333 0.06667]
          [1.      0.03061 0.04321]]]
        >>> print(chaospy.around(orth[2], 5))
        [q0^2-q0+0.16667, q1^2-0.88889q1+0.16667]
        >>> print(numpy.around(norms, 5))
        [[1.      0.08333 0.00556]
         [1.      0.03061 0.00132]]
        >>> coeffs, orth, norms = discretized_stieltjes(
        ...     2, abscissas, weights, normed=True)
        >>> print(numpy.around(coeffs, 5))
        [[[0.5     0.04167 0.26424]
          [0.42857 0.01409 0.31365]]
        <BLANKLINE>
         [[1.      1.      1.     ]
          [1.      1.      1.     ]]]
        >>> print(chaospy.around(orth[2], 5))
        [3.9155q0^2-2.1209q0-1.04874, 5.94906q1^2-2.63343q1-1.00494]
        >>> print(numpy.around(norms, 5))
        [[1. 1. 1.]
         [1. 1. 1.]]
    """
    abscissas = numpy.asfarray(abscissas)
    weights = numpy.asfarray(weights)
    assert len(weights.shape) == 1
    assert len(abscissas.shape) == 2
    assert abscissas.shape[-1] == len(weights)

    poly = chaospy.poly.variable(len(abscissas))
    orth = [poly*0, poly**0]

    inner = numpy.sum(abscissas*weights, -1)
    norms = [numpy.ones(len(abscissas)), numpy.ones(len(abscissas))]
    coeffs = []

    for _ in range(int(order)):

        coeffs.append((inner/norms[-1], norms[-1]/norms[-2]))
        orth.append((poly-coeffs[-1][0])*orth[-1] - orth[-2]*coeffs[-1][1])

        raw_nodes = orth[-1](*abscissas)**2*weights
        inner = numpy.sum(abscissas*raw_nodes, -1)
        norms.append(numpy.sum(raw_nodes, -1))

        if normed:
            orth[-1] = orth[-1]/numpy.sqrt(norms[-1])
            norms[-1] **= 0

    coeffs.append((inner/norms[-1], norms[-1]/norms[-2]))

    coeffs = numpy.moveaxis(coeffs, 0, 2)
    norms = numpy.array(norms[1:]).T
    orth = chaospy.poly.Poly(orth[1:])

    return coeffs, orth, norms



def analytical_stieljes(order, dist, normed=False):
    """
    Examples:
        >>> dist = chaospy.J(chaospy.Uniform(0, 1), chaospy.Beta(3, 4))
        >>> coeffs, orth, norms = analytical_stieljes(2, dist)
        >>> print(numpy.around(coeffs, 5))
        [[[0.5     0.5     0.5    ]
          [0.42857 0.46032 0.47475]]
        <BLANKLINE>
         [[1.      0.08333 0.06667]
          [1.      0.03061 0.04321]]]
        >>> print(chaospy.around(orth[:, 2], 5))
        [q0^2-q0+0.16667, q1^2-0.88889q1+0.16667]
        >>> print(numpy.around(norms, 5))
        [[1.      0.08333 0.00556]
         [1.      0.03061 0.00132]]
        >>> coeffs, orth, norms = analytical_stieljes(2, dist, normed=True)
        >>> print(numpy.around(coeffs, 5))
        [[[0.5     0.5     0.5    ]
          [0.42857 0.46032 0.47475]]
        <BLANKLINE>
         [[1.      0.08333 0.06667]
          [1.      0.03061 0.04321]]]
        >>> print(chaospy.around(orth[:, 2], 5))
        [13.41641q0^2-13.41641q0+2.23607, 27.49545q1^2-24.4404q1+4.58258]
        >>> print(numpy.around(norms, 5))
        [[1. 1. 1.]
         [1. 1. 1.]]
    """
    dimensions = len(dist)
    mom_order = numpy.arange(order+1).repeat(dimensions)
    mom_order = mom_order.reshape(order+1, dimensions).T
    coeffs = dist.ttr(mom_order)
    coeffs[1, :, 0] = 1.

    var = chaospy.poly.collection.core.variable(dimensions)
    orth = [var-var, var**0*numpy.ones(dimensions)]
    for order_ in range(order):
        orth.append(
            orth[-1]*(var-coeffs[0, :, order_])
            - orth[-2]*coeffs[1, :, order_]
        )
    orth = chaospy.poly.transpose(chaospy.poly.Poly(orth[1:]))

    norms = numpy.cumprod(coeffs[1], 1)
    if normed:
        orth /= numpy.sqrt(norms)
        norms **= 0

    return coeffs, orth, norms



"""Uniform probability distribution."""
from ..baseclass import Dist
from ..operators.addition import Add


class uniform(Dist):
    """Uniform distribution fixed on the [-1, 1] interval."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return 0.5

    def _cdf(self, x):
        return .5*x+.5

    def _ppf(self, q):
        return 2*q-1

    def _bnd(self, x):
        return -1.,1.

    def _mom(self, k):
        return 1./(k + 1)* (k % 2 == 0)

    def _ttr(self, n):
        return 0., n*n/(4.*n*n-1)


class Uniform(Add):
    r"""
    Uniform probability distribution.

    Args:
        lower (float, Dist): Lower threshold of distribution. Must be smaller
            than ``upper``.
        upper (float, Dist): Upper threshold of distribution.

    Examples:
        >>> distribution = chaospy.Uniform(2, 4)
        >>> print(distribution)
        Uniform(lower=2, upper=4)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.  2.5 3.  3.5 4. ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.5 0.5 0.5 0.5 0.5]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.3072 2.23   3.9006 2.9644]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[3.     3.     3.    ]
         [0.3333 0.2667 0.2571]]
    """

    def __init__(self, lower=0, upper=1):
        self._repr = {"lower": lower, "upper": upper}
        left = uniform()*((upper-lower)*.5)
        right = 0.5*(upper+lower)
        Add.__init__(self, left=left, right=right)



"""
As the number of dimensions increases linear, the number of samples increases
exponentially. This is known as the curse of dimensionality. Except for
switching to Monte Carlo integration, the is no way to completly guard against
this problem. However, there are some possibility to mitigate the problem
personally. One such strategy is to employ Smolyak sparse-grid quadrature. This
method uses a quadrature rule over a combination of different orders to tailor
a scheme that uses fewer abscissas points than a full tensor-product approach.

To use Smolyak sparse-grid in ``chaospy``, just pass the flag ``sparse=True``
to the ``generate_quadrature`` function. For example::

    >>> distribution = chaospy.J(
    ...     chaospy.Uniform(0, 4), chaospy.Uniform(0, 4))
    >>> X, W = chaospy.generate_quadrature(3, distribution, sparse=True)
    >>> print(numpy.around(X, 4))
    [[0. 0. 0. 1. 2. 2. 2. 2. 2. 3. 4. 4. 4.]
     [0. 2. 4. 2. 0. 1. 2. 3. 4. 2. 0. 2. 4.]]
    >>> print(numpy.around(W, 4))  # doctest: +NORMALIZE_WHITESPACE
    [-0.0833  0.2222 -0.0833  0.4444  0.2222  0.4444 -1.3333
      0.4444  0.2222  0.4444 -0.0833  0.2222 -0.0833]

This compared to the full tensor-product grid::

    >>> X, W = chaospy.generate_quadrature(3, distribution, sparse=False)
    >>> print(numpy.around(X, 4))
    [[0. 0. 0. 0. 1. 1. 1. 1. 3. 3. 3. 3. 4. 4. 4. 4.]
     [0. 1. 3. 4. 0. 1. 3. 4. 0. 1. 3. 4. 0. 1. 3. 4.]]
    >>> print(numpy.around(W, 4))  # doctest: +NORMALIZE_WHITESPACE
    [0.0031 0.0247 0.0247 0.0031 0.0247 0.1975 0.1975 0.0247
     0.0247 0.1975 0.1975 0.0247 0.0031 0.0247 0.0247 0.0031]

The method works with all quadrature rules, but is known to be quite
inefficient when applied to rules that can not be nested. For example using
Gauss-Legendre samples::

    >>> X, W = chaospy.generate_quadrature(
    ...     6, distribution, rule="gauss_legendre", sparse=True)
    >>> print(len(W))
    140
    >>> X, W = chaospy.generate_quadrature(
    ...     6, distribution, rule="gauss_legendre", sparse=False)
    >>> print(len(W))
    49

.. note:
    Some quadrature rules are only partially nested at certain orders. These
    include e.g. :ref:`clenshaw_curtis`, :ref:`fejer` and :ref:`newton_cotes`.
    To exploit this nested-nes, the default behavior is to only include orders
    that are properly nested. This implies that flipping the flag ``sparse``
    will result in a somewhat different scheme. To fix the scheme one way or
    the other, explicitly include the flag ``growth=False`` or ``growth=True``
    respectively.
"""
from collections import defaultdict
from itertools import product

import numpy
from scipy.special import comb

from ..bertran import bindex


def construct_sparse_grid(
        order,
        dist,
        rule="gaussian",
        accuracy=100,
        growth=None,
        recurrence_algorithm="",
):
    """
    Smolyak sparse grid constructor.

    Args:
        order (int, numpy.ndarray):
            The order of the grid. If ``numpy.ndarray``, it overrides both
            ``dim`` and ``skew``.
        dist (chaospy.distributions.baseclass.Dist):
            The distribution which density will be used as weight function.
        rule (str):
            Rule for generating abscissas and weights. Either done with
            quadrature rules, or with random samples with constant weights.
        accuracy (int):
            If gaussian is set, but the Dist provieded in domain does not
            provide an analytical TTR, ac sets the approximation order for the
            descitized Stieltje's method.
        growth (bool, None):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples. Defaults to the same
            value as ``sparse`` if omitted.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights in
            case of Gaussian quadrature scheme. If omitted, ``analytical`` will
            be tried first, and ``stieltjes`` used if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            Abscissas and weights created from sparse grid rule. Flatten such
            that ``abscissas.shape == (len(dist), len(weights))``.

    Example:
        >>> distribution = chaospy.J(
        ...     chaospy.Normal(0, 1), chaospy.Uniform(-1, 1))
        >>> X, W = construct_sparse_grid(1, distribution)
        >>> print(numpy.around(X, 4))
        [[-1.      0.      0.      0.      1.    ]
         [ 0.     -0.5774  0.      0.5774  0.    ]]
        >>> print(numpy.around(W, 4))
        [ 0.5  0.5 -1.   0.5  0.5]
        >>> X, W = construct_sparse_grid([2, 1], distribution)
        >>> print(numpy.around(X, 3))
        [[-1.732 -1.    -1.    -1.     0.     1.     1.     1.     1.732]
         [ 0.    -0.577  0.     0.577  0.    -0.577  0.     0.577  0.   ]]
        >>> print(numpy.around(W, 3))
        [ 0.167  0.25  -0.5    0.25   0.667  0.25  -0.5    0.25   0.167]
    """
    orders = order*numpy.ones(len(dist), dtype=int)

    if isinstance(rule, str):
        rule = (rule,)*len(dist)

    x_lookup, w_lookup = _construct_lookup(
        orders, dist, rule, accuracy, growth, recurrence_algorithm)
    collection = _construct_collection(
        order, dist, x_lookup, w_lookup)

    abscissas = sorted(collection)
    weights = numpy.array([collection[key] for key in abscissas])
    abscissas = numpy.array(abscissas).T
    return abscissas, weights


def _construct_collection(
        orders,
        dist,
        x_lookup,
        w_lookup,
):
    """Create a collection of {abscissa: weight} key-value pairs."""
    order = numpy.min(orders)
    skew = orders-order

    # Indices and coefficients used in the calculations
    indices = bindex(order-len(dist)+1, order, dim=len(dist))
    coeffs = numpy.sum(indices, -1)
    coeffs = (2*((order-coeffs+1) % 2)-1)*comb(len(dist)-1, order-coeffs)

    collection = defaultdict(float)
    for bidx, coeff in zip(indices+skew, coeffs.tolist()):
        abscissas = [value[idx] for idx, value in zip(bidx, x_lookup)]
        weights = [value[idx] for idx, value in zip(bidx, w_lookup)]
        for abscissa, weight in zip(product(*abscissas), product(*weights)):
            collection[abscissa] += numpy.prod(weight)*coeff

    return collection


def _construct_lookup(
        orders,
        dists,
        rules,
        accuracy,
        growth,
        recurrence_algorithm,
):
    """
    Create abscissas and weights look-up table so values do not need to be
    re-calculatated on the fly.
    """
    from .frontend import generate_quadrature
    x_lookup = []
    w_lookup = []
    for max_order, dist, rule in zip(orders, dists, rules):
        x_lookup.append([])
        w_lookup.append([])
        for order in range(max_order+1):
            (abscissas,), weights = generate_quadrature(
                order,
                dist,
                accuracy=accuracy,
                rule=rule,
                growth=growth,
                recurrence_algorithm=recurrence_algorithm,
            )
            x_lookup[-1].append(abscissas)
            w_lookup[-1].append(weights)
    return x_lookup, w_lookup



"""Convert three terms recurrence coefficients into quadrature rules."""
import numpy
import scipy.linalg


def coefficients_to_quadrature(coeffs):
    """
    Construct Gaussian quadrature abscissas and weights from three terms
    recurrence coefficients.

    Examples:
        >>> distribution = chaospy.Normal(0, 1)
        >>> coeffs = chaospy.construct_recurrence_coefficients(4, distribution)
        >>> print(numpy.around(coeffs, 4))
        [[[0. 0. 0. 0. 0.]
          [0. 1. 2. 3. 4.]]]
        >>> abscissas, weights = chaospy.coefficients_to_quadrature(coeffs)
        >>> print(numpy.around(abscissas, 4))
        [[-2.857  -1.3556 -0.      1.3556  2.857 ]]
        >>> print(numpy.around(weights, 4))
        [[0.0113 0.2221 0.5333 0.2221 0.0113]]
    """
    coeffs = numpy.asfarray(coeffs)
    if len(coeffs.shape) == 2:
        coeffs = coeffs.reshape(1, 2, -1)
    assert len(coeffs.shape) == 3, "shape %s not allowed" % coeffs.shape
    assert coeffs.shape[-1] >= 1
    abscissas = []
    weights = []
    for coeff in coeffs:

        if numpy.any(coeff[1] < 0) or numpy.any(numpy.isnan(coeff)):
            raise numpy.linalg.LinAlgError(
                "Invalid recurrence coefficients can not be used for "
                "constructing Gaussian quadrature rule")

        order = len(coeff[0])
        if order:
            bands = numpy.zeros((2, order))
            bands[0, :] = coeff[0, :order]
            bands[1, :-1] = numpy.sqrt(coeff[1, 1:order])
            vals, vecs = scipy.linalg.eig_banded(bands, lower=True)

            abscissa, weight = vals.real, vecs[0, :]**2
            indices = numpy.argsort(abscissa)
            abscissa, weight = abscissa[indices], weight[indices]

        else:
            abscissa, weight = numpy.array([coeffs[dim][0, 0]]), numpy.array([1.])

        abscissas.append(abscissa)
        weights.append(weight)

    return abscissas, weights



"""Fatigue-life distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class fatigue_life(Dist):
    """Fatigue-life distribution."""

    def __init__(self, c=0):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        output = (x+1)/(2*c*numpy.sqrt(2*numpy.pi*x**3))
        output *= numpy.exp(-(x-1)**2/(2.0*x*c**2))
        output[(x == 0) & numpy.isnan(output)] = 0
        return output

    def _cdf(self, x, c):
        return special.ndtr(1.0/c*(numpy.sqrt(x)-1.0/numpy.sqrt(x)))

    def _ppf(self, q, c):
        tmp = c*special.ndtri(q)
        return 0.25*(tmp + numpy.sqrt(tmp**2 + 4))**2

    def _bnd(self, x, c):
        return 0, self._ppf(1-1e-10, c)


class FatigueLife(Add):
    """
    Fatigue-Life or Birmbaum-Sanders distribution

    Args:
        shape (float, Dist): Shape parameter
        scale (float, Dist): Scaling parameter
        shift (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.FatigueLife(2, 2, 1)
        >>> print(distribution)
        FatigueLife(scale=2, shape=2, shift=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [ 1.4332  2.2113  4.3021 10.2334]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.4223 0.1645 0.0603 0.0198]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 5.3231  1.2621 26.5603  2.8292]
        >>> print(numpy.around(distribution.mom(1), 4))
        7.0
    """
    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=fatigue_life(shape)*scale, right=shift)


Fatiguelife = deprecation_warning(FatigueLife, "Fatiguelife")



"""Multivariate Student-T Distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from .student_t import student_t
from .deprecate import deprecation_warning


class MvStudentT(Dist):
    """
    Multivariate Student-T Distribution.

    Args:
        df (float, Dist) : Degree of freedom
        loc (numpy.ndarray, Dist) : Location parameter
        scale (numpy.ndarray) : Covariance matrix

    Examples:
        >>> distribution = chaospy.MvStudentT(40, [1, 2], [[1, 0.6], [0.6, 1]])
        >>> print(distribution)
        MvStudentT(df=40, loc=[1.0, 2.0], scale=[[1.0, 0.6], [0.6, 1.0]])
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(distribution.inv(mesh), 4))
        [[[0.3193 1.     1.6807]
          [0.3193 1.     1.6807]
          [0.3193 1.     1.6807]]
        <BLANKLINE>
         [[1.0471 1.4555 1.8639]
          [1.5916 2.     2.4084]
          [2.1361 2.5445 2.9529]]]
        >>> print(numpy.around(distribution.fwd(distribution.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(distribution.pdf(distribution.inv(mesh)), 4))
        [[0.1921 0.1959 0.1958]
         [0.197  0.1989 0.197 ]
         [0.1958 0.1959 0.1921]]
        >>> print(numpy.around(distribution.sample(4), 4))
        [[ 1.3979 -0.2189  2.6868  0.9551]
         [ 3.1625  0.6234  1.582   1.7631]]
    """

    def __init__(self, df=1, loc=[0, 0], scale=[[1, .5], [.5, 1]]):
        loc = numpy.asfarray(loc)
        scale = numpy.asfarray(scale)
        assert len(loc) == len(scale)
        self.length = len(loc)
        self._repr = {"df": df, "loc": loc.tolist(), "scale": scale.tolist()}

        C = numpy.linalg.cholesky(scale)
        Ci = numpy.linalg.inv(C)
        Dist.__init__(self, a=df, C=C, Ci=Ci, loc=loc)

    def _cdf(self, x, a, C, Ci, loc):
        x = numpy.dot(Ci, (x.T-loc.T).T)
        return special.stdtr(a, x)

    def _ppf(self, q, a, C, Ci, loc):
        z = special.stdtrit(a, q)
        out = (numpy.dot(C, z).T + loc.T).T
        return out

    def _pdf(self, x, a, C, Ci, loc):

        det = numpy.linalg.det(numpy.dot(C,C.T))
        k = len(C)

        x_ = numpy.dot(Ci.T, (x.T-loc.T).T)
        out = numpy.ones(x.shape)
        out[0] = special.gamma(.5*(a+k))/(special.gamma(.5*a)* \
                a**(.5*k)*numpy.pi**(.5*k)*det**.5*\
                (1+numpy.sum(x_*x_,0)/a))
        return out

    def _bnd(self, x, a, C, Ci, loc):
        output = numpy.zeros((2,)+x.shape)
        scale = numpy.sqrt(numpy.diag(numpy.dot(C, C.T)))
        output.T[:, :, 0] = -10**3*scale + loc
        output.T[:, :, 1] = 10**3*scale + loc
        return output

    def __len__(self):
        return self.length

    # def _mom(self, k, a, C, Ci, loc):

    #     scale = numpy.dot(C, C.T)

    #     def mom(k):

    #         zeros = (numpy.sum(k,0)%2==1)+numpy.any(numpy.array(k)<0, 0)
    #         if numpy.all(zeros, 0):
    #             return 0.

    #         dim, K = k.shape
    #         ra = numpy.arange(dim).repeat(K).reshape(dim,K)

    #         i = numpy.argmax(k!=0, 0)

    #         out = numpy.zeros(k.shape[1:])
    #         out[:] = numpy.where(numpy.choose(i,k),
    #                 (numpy.choose(i,k)-1)*scale[i,i]*mom(k-2*(ra==i)), 1)
    #         for x in range(1, dim):
    #             out += \
    #             (numpy.choose(i,k)!=0)*(x>i)*k[x]*scale[i,x]*mom(k-(ra==i)-(ra==x))

    #         return out

    #     dim = len(loc)
    #     K = numpy.mgrid[[slice(0,_+1,1) for _ in numpy.max(k, 1)]]
    #     K = K.reshape(dim, K.size//dim)
    #     M = mom(K)

    #     out = numpy.zeros(k.shape[1])
    #     for i in range(len(M)):
    #         coef = numpy.prod(special.comb(k.T, K[:,i]).T, 0)
    #         diff = k.T - K[:,i]
    #         pos = diff>=0
    #         diff = diff*pos
    #         pos = numpy.all(pos, 1)
    #         loc_ = numpy.prod(loc**diff, 1)
    #         out += pos*coef*loc_*M[i]

    #     return out


MvStudent_t = deprecation_warning(MvStudentT, "MvStudent_t")



"""Give warning for using old names."""
import logging
from functools import wraps


def deprecation_warning(func, name):
    """Add a deprecation warning do each distribution."""
    @wraps(func)
    def caller(*args, **kwargs):
        """Docs to be replaced."""
        logger = logging.getLogger(__name__)
        instance = func(*args, **kwargs)
        logger.warning(
            "Distribution `chaospy.{}` has been renamed to ".format(name) +
            "`chaospy.{}` and will be deprecated next release.".format(instance.__class__.__name__))
        return instance
    return caller



"""
Constructing a multivariate probability distribution consisting of identical
independent distributed marginals can be done using the
:func:`~chaospy.distributions.operators.iid.Iid`. E.g.::

    >>> X = chaospy.Normal()
    >>> Y = chaospy.Iid(X, 4)
    >>> print(Y.sample())
    [ 0.39502989 -1.20032309  1.64760248 -0.04465437]
"""
import numpy

from ..baseclass import Dist
from .. import evaluation


class Iid(Dist):
    """
    Opaque method for creating independent identical distributed random
    variables from an univariate variable.

    Args:
        dist (Dist):
            Distribution to make into i.i.d. vector.
    """

    def __init__(self, dist, length):
        assert len(dist) == 1 and length >= 1
        Dist.__init__(self, dist=dist, length=length)

    def _pdf(self, xloc, dist, length, cache):
        """
        Probability density function.

        Example:
            >>> print(chaospy.Iid(chaospy.Uniform(), 2).pdf(
            ...     [[0.5, 1.5], [0.5, 0.5]]))
            [1. 0.]
        """
        output = evaluation.evaluate_density(
            dist, xloc.reshape(1, -1)).reshape(length, -1)
        assert xloc.shape == output.shape
        return output

    def _cdf(self, xloc, dist, length, cache):
        """
        Cumulative distribution function.

        Example:
            >>> print(chaospy.Iid(chaospy.Uniform(0, 2), 2).fwd(
            ...     [[0.1, 0.2, 0.3], [0.2, 0.2, 0.3]]))
            [[0.05 0.1  0.15]
             [0.1  0.1  0.15]]
        """
        output = evaluation.evaluate_forward(
            dist, xloc.reshape(1, -1)).reshape(length, -1)
        assert xloc.shape == output.shape
        return output

    def _ppf(self, uloc, dist, length, cache):
        """
        Point percentile function.

        Example:
            >>> print(chaospy.Iid(chaospy.Uniform(0, 2), 2).inv(
            ...     [[0.1, 0.2, 0.3], [0.2, 0.2, 0.3]]))
            [[0.2 0.4 0.6]
             [0.4 0.4 0.6]]
        """
        output = evaluation.evaluate_inverse(
            dist, uloc.reshape(1, -1)).reshape(length, -1)
        assert uloc.shape == output.shape
        return output

    def _bnd(self, xloc, dist, length, cache):
        """
        boundary function.

        Example:
            >>> print(chaospy.Iid(chaospy.Uniform(0, 2), 2).range(
            ...     [[0.1, 0.2, 0.3], [0.2, 0.2, 0.3]]))
            [[[0. 0. 0.]
              [0. 0. 0.]]
            <BLANKLINE>
             [[2. 2. 2.]
              [2. 2. 2.]]]
        """
        lower, upper = evaluation.evaluate_bound(
            dist, xloc.reshape(1, -1))
        lower = lower.reshape(length, -1)
        upper = upper.reshape(length, -1)
        assert lower.shape == xloc.shape, (lower.shape, xloc.shape)
        assert upper.shape == xloc.shape
        return lower, upper

    def _mom(self, k, dist, length, cache):
        """
        Moment generating function.

        Example:
            >>> print(chaospy.Iid(chaospy.Uniform(), 2).mom(
            ...     [[0, 0, 1], [0, 1, 1]]))
            [1.   0.5  0.25]
        """
        return numpy.prod(dist.mom(k), 0)

    def _ttr(self, k, dist, length, cache):
        """
        Three terms recursion generating function.

        Example:
            >>> print(numpy.around(chaospy.Iid(chaospy.Uniform(), 2).ttr(
            ...     [[0, 0, 1], [0, 1, 1]]), 4))
            [[[ 0.5     0.5     0.5   ]
              [ 0.5     0.5     0.5   ]]
            <BLANKLINE>
             [[-0.     -0.      0.0833]
              [-0.      0.0833  0.0833]]]
        """
        coeffs = [dist.ttr(k_).T for k_ in k]
        coeffs = list(zip(*coeffs))
        return coeffs

    def __getitem__(self, i):
        """
        Slicing function.

        Example:
            >>> dist = chaospy.Iid(chaospy.Uniform(), 3)
            >>> print(dist[0])
            Uniform(lower=0, upper=1)
            >>> print(dist[1])
            Uniform(lower=0, upper=1)
            >>> print(dist[:1])
            Iid(Uniform(lower=0, upper=1), 1)
            >>> print(dist[1:])
            Iid(Uniform(lower=0, upper=1), 2)
        """
        if isinstance(i, int):
            if (i >= len(self)) or (i < -len(self)):
                raise IndexError("dist index out of range")
            return self.prm["dist"]
        if isinstance(i, slice):
            start, stop, step = i.start, i.stop, i.step
            if start is None: start = 0
            if stop is None: stop = len(self)
            if step is None: step = 1
            length = (stop - start) // step
            return Iid(self.prm["dist"], length=length)
        raise IndexError("index not recognised.")

    def __len__(self):
        """
        Example:
            >>> len(chaospy.Iid(chaospy.Uniform(), 4))
            4
            >>> len(chaospy.Iid(chaospy.Uniform(), 9))
            9
        """
        return self.prm["length"]

    def __str__(self):
        return (self.__class__.__name__ + "(" + str(self.prm["dist"]) +
                ", " + str(self.prm["length"]) + ")")



"""Generalized exponential distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class generalized_exponential(Dist):
    """Generalized exponential distribution."""

    def __init__(self, a=1, b=1, c=1):
        Dist.__init__(self, a=a, b=b, c=c)

    def _pdf(self, x, a, b, c):
        return (a+b*(-numpy.expm1(-c*x)))*numpy.exp((-a-b)*x+b*(-numpy.expm1(-c*x))/c)

    def _cdf(self, x, a, b, c):
        output = -numpy.expm1((-a-b)*x + b*(-numpy.expm1(-c*x))/c)
        output = numpy.where(x > 0, output, 0)
        return output

    def _bnd(self, x, a, b, c):
        return 0, 10**10


class GeneralizedExponential(Add):
    """
    Generalized exponential distribution.

    Args:
        a (float, Dist) : First shape parameter
        b (float, Dist) : Second shape parameter
        c (float, Dist) : Third shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Note:
        "An Extension of Marshall and Olkin's Bivariate Exponential Distribution",
        H.K. Ryu, Journal of the American Statistical Association, 1993.

        "The Exponential Distribution: Theory, Methods and Applications",
        N. Balakrishnan, Asit P. Basu.

    Examples:
        >>> distribution = chaospy.GeneralizedExponential(3, 2, 2, 2, 2)
        >>> print(distribution)
        GeneralizedExponential(a=3, b=2, c=2, scale=2, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.1423 2.3113 2.5314 2.8774]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.3061 1.0605 0.7649 0.4168]
        >>> print(numpy.around(distribution.sample(4), 4))
        [2.6058 2.0794 3.5124 2.3933]
    """

    def __init__(self, a=1, b=1, c=1, scale=1, shift=0):
        self._repr = {"a": a, "b": b, "c": c, "scale": scale, "shift": shift}
        Add.__init__(
            self, left=generalized_exponential(a, b, c)*scale, right=shift)


Genexpon = deprecation_warning(GeneralizedExponential, "Genexpon")



"""
Gauss-Radau formula for numerical estimation of integrals. It requires
:math:`m+1` points and fits all Polynomials to degree :math:`2m`, so it
effectively fits exactly all Polynomials of degree :math:`2m-1`.

It allows for a single abscissas to be user defined, while the others are built
around this point.

Canonically, Radau is built around Legendre weight function with the fixed
point at the left end. Not all distributions/fixed point combinations allows
for the building of a quadrature scheme.

Example usage
-------------

With increasing order::

    >>> distribution = chaospy.Beta(2, 2, lower=-1, upper=1)
    >>> for order in range(4):  # doctest: +NORMALIZE_WHITESPACE
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="gauss_radau")
    ...     print(numpy.around(X, 2), numpy.around(W, 2))
    [[-1.]] [1.]
    [[-1.   0.2]] [0.17 0.83]
    [[-1.   -0.51  0.13  0.71]] [0.02 0.33 0.48 0.17]
    [[-1.   -0.74 -0.35  0.1   0.53  0.85]]
     [0.01 0.11 0.28 0.34 0.21 0.05]

Multivariate samples::

    >>> distribution = chaospy.J(
    ...     chaospy.Uniform(0, 1), chaospy.Beta(4, 5))
    >>> X, W = chaospy.generate_quadrature(
    ...     1, distribution, rule="gauss_radau")
    >>> print(numpy.around(X, 3))
    [[0.    0.    0.667 0.667]
     [0.    0.5   0.    0.5  ]]
    >>> print(numpy.around(W, 3))
    [0.028 0.222 0.083 0.667]

To change the fixed point, the direct generating function has to be used::

    >>> distribution = chaospy.Uniform(lower=-1, upper=1)
    >>> for fixed_point in numpy.linspace(-1, 1, 6):
    ...     X, W = chaospy.quad_gauss_radau(2, distribution, fixed_point)
    ...     print(numpy.around(X, 2), numpy.around(W, 2))
    [[-1.   -0.58  0.18  0.82]] [0.06 0.33 0.39 0.22]
    [[-1.04 -0.6   0.17  0.82]] [0.05 0.33 0.39 0.22]
    [[-0.83 -0.2   0.54  0.96]] [0.22 0.38 0.32 0.08]
    [[-0.96 -0.54  0.2   0.83]] [0.08 0.32 0.38 0.22]
    [[-0.82 -0.17  0.6   1.04]] [0.22 0.39 0.33 0.05]
    [[-0.82 -0.18  0.58  1.  ]] [0.22 0.39 0.33 0.06]

However, a fixed point at 0 is not allowed::

    >>> chaospy.quad_gauss_radau(  # doctest: +IGNORE_EXCEPTION_DETAIL
    ...     3, distribution, fixed_point=0)
    Traceback (most recent call last):
        ...
    numpy.linalg.LinAlgError: Illegal Radau fixed point: 0.0
"""
from __future__ import print_function

import numpy
import scipy.linalg

from .recurrence import (
    construct_recurrence_coefficients, coefficients_to_quadrature)
from .combine import combine_quadrature


def quad_gauss_radau(
        order,
        dist,
        fixed_point=None,
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    """
    Generate the quadrature nodes and weights in Gauss-Radau quadrature.

    Args:
        order (int):
            Quadrature order.
        dist (chaospy.distributions.baseclass.Dist):
            The distribution weights to be used to create higher order nodes
            from.
        fixed_point (float):
            Fixed point abscissas assumed to be included in the quadrature. If
            imitted, use distribution lower point ``dist.range()[0]``.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        accuracy (int):
            In the case ``rule`` is used, defines the quadrature order of the
            scheme used. In practice, must be at least as large as ``order``.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Raises:
        ValueError:
            Error raised if Radau algorithm fails to find recurrence
            coefficients.

    Example:
        >>> abscissas, weights = quad_gauss_radau(4, chaospy.Uniform(-1, 1))
        >>> print(numpy.around(abscissas, 3))
        [[-1.    -0.887 -0.64  -0.295  0.094  0.468  0.771  0.955]]
        >>> print(numpy.around(weights, 3))
        [0.016 0.093 0.152 0.188 0.196 0.174 0.125 0.057]
    """
    assert not rule.startswith("gauss"), "recursive Gaussian quadrature call"
    if fixed_point is None:
        fixed_point, _ = dist.range()
    else:
        fixed_point = numpy.ones(len(dist))*fixed_point

    if order == 0:
        return fixed_point.reshape(-1, 1), numpy.ones(1)

    coefficients = construct_recurrence_coefficients(
        2*order-1, dist, rule, accuracy, recurrence_algorithm)
    coefficients = [radau_jakobi(coeffs, point)
                    for point, coeffs in zip(fixed_point, coefficients)]

    abscissas, weights = coefficients_to_quadrature(coefficients)

    return combine_quadrature(abscissas, weights)


def radau_jakobi(coeffs, fixed_point):
    """
    Compute the Radau coefficients.

    Based on the section 7 of the paper "Some modified matrix eigenvalue
    problems", G. Golub.

    Args:
        coeffs_a (numpy.ndarray):
            The first three terms recurrence coefficients of the Gaussian
            quadrature rule.
        coeffs_b (numpy.ndarray):
            The second three terms recurrence coefficients of the Gaussian
            quadrature rule.
        fixed_point (float):
            Fixed point abscissas assumed to be included in the quadrature.

    Returns:
        Three terms recurrence coefficients of the Gauss-Radau quadrature rule.

    Raises:
        numpy.linalg.LinAlgError:
            Error raised if fixed point causes the algorithm to break down.
    """
    right_hand_side = numpy.zeros(len(coeffs[0])-1)
    right_hand_side[-1] = coeffs[1][-1]
    bands_a = numpy.vstack([numpy.sqrt(coeffs[1]), coeffs[0]-fixed_point])
    bands_j = numpy.vstack((bands_a[:, 0:-1], bands_a[0, 1:]))
    try:
        delta = scipy.linalg.solve_banded((1, 1), bands_j, right_hand_side)
    except numpy.linalg.LinAlgError:
        raise numpy.linalg.LinAlgError(
            "Illegal Radau fixed point: %s" % fixed_point.item())
    coeffs = coeffs.copy()
    coeffs[0, -1] = fixed_point+delta[-1]
    return coeffs



"""Logarithm with base 10."""
import numpy

from ..baseclass import Dist
from .. import evaluation, approximation


class Log10(Dist):
    """
    Logarithm with base 10.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Log10(chaospy.Uniform(1, 2))
        >>> print(distribution)
        Log10(Uniform(lower=1, upper=2))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.0792 0.1461 0.2041 0.2553]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [2.7631 3.2236 3.6841 4.1447]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.2184 0.0473 0.2901 0.1709]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.1678
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[0.1678 0.1472 0.15  ]
         [1.     0.0074 0.0061]]
    """

    def __init__(self, dist):
        """
        Constructor.

        Args:
            dist (Dist) : distribution (>=0).
        """
        assert isinstance(dist, Dist)
        assert numpy.all(dist.range() > 0)
        Dist.__init__(self, dist=dist)

    def _pdf(self, xloc, dist, cache):
        """Probability density function."""
        return evaluation.evaluate_density(
            dist, 10**xloc, cache=cache)*10**xloc*numpy.log(10)

    def _cdf(self, xloc, dist, cache):
        """Cumulative distribution function."""
        return evaluation.evaluate_forward(dist, 10**xloc, cache=cache)

    def _ppf(self, q, dist, cache):
        """Point percentile function."""
        return numpy.log10(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, xloc, dist, cache):
        """Distribution bounds."""
        return numpy.log10(evaluation.evaluate_bound(
            dist, 10**xloc, cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.log10(dist)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return 10**dist
        return self



"""(Non-central) F distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class f(Dist):
    """F distribution."""

    def __init__(self, dfn, dfd, nc):
        Dist.__init__(self, dfn=dfn, dfd=dfd, nc=nc)

    def _pdf(self, x, dfn, dfd, nc):
        n1, n2 = dfn, dfd
        term = -nc/2.+nc*n1*x/(2*(n2+n1*x)) + special.gammaln(n1/2.)+special.gammaln(1+n2/2.)
        term -= special.gammaln((n1+n2)/2.)
        Px = numpy.exp(term)
        Px *= n1**(n1/2.) * n2**(n2/2.) * x**(n1/2.-1)
        Px *= (n2+n1*x)**(-(n1+n2)/2.)
        Px *= special.assoc_laguerre(-nc*n1*x/(2.*(n2+n1*x)), n2/2., n1/2.-1)
        Px /= special.beta(n1/2., n2/2.)
        return Px

    def _cdf(self, x, dfn, dfd, nc):
        return special.ncfdtr(dfn, dfd, nc, x)

    def _ppf(self, q, dfn, dfd, nc):
        return special.ncfdtri(dfn, dfd, nc, q)

    def _bnd(self, x, dfn, dfd, nc):
        return 0.0, self._ppf(1-1e-10, dfn, dfd, nc)


class F(Add):
    """
    (Non-central) F or Fisher-Snedecor distribution.

    Args:
        n (float, Dist) : Degres of freedom for numerator
        m (float, Dist) : Degres of freedom for denominator
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
        nc (float, Dist) : Non-centrality parameter

    Examples:
        >>> distribution = chaospy.F(3, 3, 2, 1, 1)
        >>> print(distribution)
        F(m=3, n=3, nc=1, scale=2, shift=1)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [1.9336 2.9751 4.7028 8.8521]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.2277 0.1572 0.0837 0.027 ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 5.4212  1.5739 25.7656  3.5586]
        >>> print(distribution.mom(1) > 10**8) # undefined
        True
    """

    def __init__(self, n=1, m=1, scale=1, shift=0, nc=0):
        self._repr = {"n": n, "m": m, "scale": scale, "shift": shift, "nc": nc}
        Add.__init__(self, left=f(n, m, nc)*scale, right=shift)



"""Expected value."""
import numpy

from .. import distributions, poly as polynomials, quadrature


def E(poly, dist=None, **kws):
    """
    Expected value operator.

    1st order statistics of a probability distribution or polynomial on a given
    probability space.

    Args:
        poly (Poly, Dist):
            Input to take expected value on.
        dist (Dist):
            Defines the space the expected value is taken on. It is ignored if
            ``poly`` is a distribution.

    Returns:
        (numpy.ndarray):
            The expected value of the polynomial or distribution, where
            ``expected.shape == poly.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> print(chaospy.E(dist))
        [1. 0.]
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.Poly([1, x, y, 10*x*y])
        >>> print(chaospy.E(poly, dist))
        [1. 1. 0. 0.]
    """
    if isinstance(poly, distributions.Dist):
        dist, poly = poly, polynomials.variable(len(poly))

    if not poly.keys:
        return numpy.zeros(poly.shape, dtype=int)

    if isinstance(poly, (list, tuple, numpy.ndarray)):
        return [E(_, dist, **kws) for _ in poly]

    if poly.dim < len(dist):
        poly = polynomials.setdim(poly, len(dist))

    shape = poly.shape
    poly = polynomials.flatten(poly)

    keys = poly.keys
    mom = dist.mom(numpy.array(keys).T, **kws)
    A = poly.A

    if len(dist) == 1:
        mom = mom[0]

    out = numpy.zeros(poly.shape)
    for i in range(len(keys)):
        out += A[keys[i]]*mom[i]

    out = numpy.reshape(out, shape)
    return out
